# pylint: disable=missing-docstring

import pytest
from nengo.exceptions import BuildError

from nengo_dl.builder import Builder, NengoModel, OpBuilder
from nengo_dl.tests import dummies
from nengo_dl.utils import NullProgressBar


def test_custom_builder():
    # pylint: disable=unused-variable

    class TestOp:
        sets = None
        incs = None
        reads = None
        updates = None

    ops = (TestOp(),)
    progress = NullProgressBar()

    # error if no builder registered
    with pytest.raises(BuildError, match="No registered builder"):
        Builder([ops])

    # warning if builder doesn't subclass OpBuilder
    with pytest.warns(UserWarning):

        @Builder.register(TestOp)
        class TestOpBuilder0:
            pass

    # warning when overwriting a registered builder
    with pytest.warns(UserWarning):

        @Builder.register(TestOp)
        class TestOpBuilder(OpBuilder):
            pre_built = False
            post_built = False

            def build_pre(self, signals, config):
                super().build_pre(signals, config)
                self.pre_built = True

            def build_step(self, signals):
                assert self.pre_built
                assert not self.post_built

                return 0, 1

            def build_post(self, signals):
                self.post_built = True

    builder = Builder([ops])

    builder.build_pre(signals=None, config=None, progress=progress)

    result = builder.build_step(signals=None, progress=progress)

    assert len(result) == 2
    assert result[0] == 0
    assert result[1] == 1

    builder.build_post(signals=None, progress=progress)
    assert builder.op_builds[ops].post_built

    # error if builder doesn't define build_step
    @Builder.register(TestOp)
    class TestOpBuilder2(OpBuilder):
        pass

    builder = Builder([ops])
    builder.build_pre(signals=None, config=None, progress=progress)
    with pytest.raises(BuildError, match="must implement a `build_step` function"):
        builder.build_step(signals=None, progress=progress)


@pytest.mark.parametrize("fail_fast", (True, False))
def test_custom_model(fail_fast):
    model = NengoModel(fail_fast=fail_fast)

    try:
        model.add_op(dummies.Op())
    except NotImplementedError:
        assert fail_fast
    else:
        assert not fail_fast



# pylint: disable=unused-import,ungrouped-imports

"""
Utilities to ease cross-compatibility between different versions of upstream
dependencies.
"""

import collections
import inspect

import nengo
import tensorflow as tf
from nengo._vendor.scipy.sparse import linalg_interface, linalg_onenormest
from packaging import version
from tensorflow.python.eager import context

# TensorFlow compatibility


class TFLogFilter:
    """
    Filter for cleaning up the TensorFlow log output.

    Parameters
    ----------
    err_on_deprecation : bool
        If True, treat deprecation log messages as errors.
    """

    def __init__(self, err_on_deprecation):
        self.err_on_deprecation = err_on_deprecation

    def filter(self, record):
        """
        Apply filter to log record.

        Parameters
        ----------
        record : `LogRecord`
            Message emitted by logger.

        Returns
        -------
        pass : bool
            If True, pass the ``record``, otherwise filter it out.

        Raises
        ------
        AttributeError
            If a deprecation message is detected and ``err_on_deprecation=True``.
        """

        # "tf.keras.backend.get_session is deprecated": this deprecation message
        # is raised incorrectly due to a bug, see
        # https://github.com/tensorflow/tensorflow/issues/33182
        if len(record.args) > 1 and record.args[1] == "tf.keras.backend.get_session":
            return False

        # "Output steps_run missing from loss dictionary": steps_run should
        # never have a loss defined
        if record.msg.startswith("Output steps_run missing from loss dictionary"):
            return False

        if self.err_on_deprecation and (
            "deprecation.py" in record.pathname or "deprecated" in record.msg.lower()
        ):
            msg = record.getMessage()
            raise AttributeError("Deprecation warning detected:\n%s" % msg)

        return True


tf.get_logger().addFilter(TFLogFilter(err_on_deprecation=False))

if version.parse(tf.__version__) < version.parse("2.3.0rc0"):

    def _build_map(outputs):
        """
        Vendored from ``tensorflow.python.keras.engine.functional._build_map``
        in TF 2.3.0 (with a few changes noted below).
        """
        finished_nodes = set()
        nodes_in_progress = set()
        nodes_in_decreasing_depth = []  # nodes from inputs -> outputs.
        layer_indices = {}  # layer -> in traversal order.
        for output in tf.nest.flatten(outputs):
            _build_map_helper(
                output,
                finished_nodes,
                nodes_in_progress,
                nodes_in_decreasing_depth,
                layer_indices,
            )

        # CHANGED: add `node.layer` alias for `node.outbound_layer`
        for node in nodes_in_decreasing_depth:
            node.layer = node.outbound_layer

        return nodes_in_decreasing_depth, layer_indices

    def _build_map_helper(
        tensor,
        finished_nodes,
        nodes_in_progress,
        nodes_in_decreasing_depth,
        layer_indices,
    ):
        """Recursive helper for `_build_map`."""
        layer, node_index, _ = tensor._keras_history  # pylint: disable=protected-access
        node = layer._inbound_nodes[node_index]  # pylint: disable=protected-access

        # Don't repeat work for shared subgraphs
        if node in finished_nodes:
            return

        # Prevent cycles.
        if node in nodes_in_progress:  # pragma: no cover
            raise ValueError(
                "The tensor "
                + str(tensor)
                + ' at layer "'
                + layer.name
                + '" is part of a cycle.'
            )

        # Store the traversal order for layer sorting.
        if layer not in layer_indices:
            layer_indices[layer] = len(layer_indices)

        # Propagate to all previous tensors connected to this node.
        nodes_in_progress.add(node)
        # CHANGED: `not node.is_input` to `node.inbound_layers`
        if node.inbound_layers:
            # CHANGED: `node.keras_inputs` to `tf.nest.flatten(node.input_tensors)`
            # CHANGED: `tensor` to `input_tensor` (for pylint)
            for input_tensor in tf.nest.flatten(node.input_tensors):
                _build_map_helper(
                    input_tensor,
                    finished_nodes,
                    nodes_in_progress,
                    nodes_in_decreasing_depth,
                    layer_indices,
                )

        finished_nodes.add(node)
        nodes_in_progress.remove(node)
        nodes_in_decreasing_depth.append(node)


else:
    from tensorflow.python.keras.engine.functional import _build_map


if version.parse(tf.__version__) < version.parse("2.3.0rc0"):
    from tensorflow.python.keras.engine import network

    # monkeypatch to fix bug in TF2.2, see
    # https://github.com/tensorflow/tensorflow/issues/37548
    old_conform = network.Network._conform_to_reference_input

    def _conform_to_reference_input(self, tensor, ref_input):
        keras_history = getattr(tensor, "_keras_history", None)

        tensor = old_conform(self, tensor, ref_input)

        if keras_history is not None:
            tensor._keras_history = keras_history

        return tensor

    network.Network._conform_to_reference_input = _conform_to_reference_input


# Nengo compatibility

if version.parse(nengo.__version__) < version.parse("3.1.0"):
    PoissonSpiking = RegularSpiking = StochasticSpiking = Tanh = NoTransform = type(
        None
    )

    default_transform = 1

    def conn_has_weights(conn):
        """All connections have weights."""
        return True

    def neuron_state(neuron_op):
        """Look up keys from function signature."""
        names = list(inspect.signature(neuron_op.neurons.step_math).parameters.keys())[
            3:
        ]
        assert len(names) == len(neuron_op.states)
        return collections.OrderedDict((n, s) for n, s in zip(names, neuron_op.states))

    def neuron_step(neuron_op, dt, J, output, state):  # pragma: no cover (runs in TF)
        """Call step_math instead of step."""
        neuron_op.neurons.step_math(dt, J, output, *state.values())

    def to_neurons(conn):
        """Check whether the output of a connection is a neuron object."""
        return isinstance(conn.post_obj, nengo.ensemble.Neurons) or (
            isinstance(conn.pre_obj, nengo.Ensemble)
            and isinstance(conn.post_obj, nengo.Ensemble)
            and conn.solver.weights
        )

    class SimProbe(nengo.builder.Operator):
        """Backport of Nengo 3.1.0 SimProbe."""

        def __init__(self, signal, tag=None):
            super().__init__(tag=tag)
            self.sets = []
            self.incs = []
            self.reads = [signal]
            self.updates = []

        def make_step(self, signals, dt, rng):
            """Not used in NengoDL."""
            raise NotImplementedError

    from nengo_dl.builder import NengoBuilder as _NengoBuilder

    def build_probe(model, probe):
        """Copy of the base probe builder that adds the SimProbe op."""
        nengo.builder.probe.build_probe(model, probe)
        model.add_op(SimProbe(model.sig[probe]["in"]))

    _NengoBuilder.register(nengo.Probe)(build_probe)

    # monkeypatch fix for https://github.com/nengo/nengo/pull/1587
    linalg_onenormest.aslinearoperator = linalg_interface.aslinearoperator

else:
    from nengo.builder.probe import SimProbe
    from nengo.neurons import PoissonSpiking, RegularSpiking, StochasticSpiking, Tanh
    from nengo.transforms import NoTransform

    default_transform = None

    def conn_has_weights(conn):
        """Equivalent to conn.has_weights."""
        return conn.has_weights

    def neuron_state(neuron_op):
        """Equivalent to neuron_op.state."""
        return neuron_op.state

    def neuron_step(neuron_op, dt, J, output, state):  # pragma: no cover (runs in TF)
        """Equivalent to neuron_op.step."""
        neuron_op.neurons.step(dt, J, output, **state)

    def to_neurons(conn):
        """Equivalent to conn._to_neurons."""
        return conn._to_neurons


if version.parse(nengo.__version__) <= version.parse("3.1.0"):

    class FrozenOrderedSet(collections.abc.Set):
        """Backport of `nengo.utils.stdlib.FrozenOrderedSet`."""

        def __init__(self, data):
            self.data = collections.OrderedDict((d, None) for d in data)

        def __contains__(self, elem):
            return elem in self.data

        def __iter__(self):
            return iter(self.data)

        def __len__(self):
            return len(self.data)

        def __hash__(self):
            return self._hash()


else:
    from nengo.utils.stdlib import FrozenOrderedSet


def eager_enabled():
    """
    Check if we're in eager mode or graph mode.

    Note: this function differs from ``tf.executing_eagerly()`` in that it will still
    return ``True`` if we're inside a ``tf.function``. Essentially this checks whether
    the user has called ``tf.compat.v1.disable_eager_execution()`` or not.
    """

    return context.default_execution_mode == context.EAGER_MODE


try:
    from keras_spiking import Alpha, Lowpass, SpikingActivation
except ImportError:
    SpikingActivation = object()
    Lowpass = object()
    Alpha = object()



"""Dummy objects for use in tests."""

from collections import defaultdict

import nengo
import numpy as np
from packaging import version

from nengo_dl import builder, signals, tensor_graph


class Signal(nengo.builder.signal.Signal):
    """
    Mock-up for `nengo.builder.Signal`.
    """

    def __init__(
        self,
        shape=None,
        dtype=None,
        base_shape=None,
        offset=0,
        trainable=False,
        label="",
        initial_value=0,
        sparse=False,
    ):  # pylint: disable=super-init-not-called
        self._shape = (1,) if shape is None else shape
        self._dtype = np.float32 if dtype is None else dtype
        self._base = (
            self
            if base_shape is None
            else Signal(shape=base_shape, dtype=self.dtype, label="%s.base" % label)
        )
        self._elemoffset = offset
        self.name = label
        self._ndim = len(self.shape)
        self._is_view = base_shape is not None
        self._size = np.prod(self.shape)
        self.trainable = trainable
        self.minibatched = not trainable
        self.init = initial_value
        self._sparse = sparse

    @property
    def initial_value(self):
        """Initial signal value."""
        return np.full(self.base.shape, self.init, dtype=self.dtype)

    @property
    def shape(self):
        """Signal shape."""
        return self._shape

    @property
    def dtype(self):
        """Signal dtype."""
        return self._dtype

    @property
    def elemoffset(self):
        """Signal offset."""
        return self._elemoffset

    @property
    def ndim(self):
        """Number of dimensions."""
        return self._ndim

    @property
    def is_view(self):
        """Whether or not this signal is a view of a base signal."""
        return self._is_view

    @property
    def size(self):
        """Size of signal."""
        return self._size

    @property
    def sparse(self):
        """Whether or not signal is sparse."""
        return self._sparse

    def may_share_memory(self, _):
        """Whether or not this signal shares memory with another."""
        return False

    def __repr__(self):
        return "DummySignal(%s)" % self.name


class Op:
    """
    Mock-up for `nengo.builder.Operator`.
    """

    def __init__(self, sets=None, incs=None, reads=None, updates=None):
        self.sets = [] if sets is None else sets
        self.incs = [] if incs is None else incs
        self.reads = [] if reads is None else reads
        self.updates = [] if updates is None else updates

        self.all_signals = self.sets + self.incs + self.reads + self.updates

    def __repr__(self):
        rep = "DummyOp("
        if len(self.sets) > 0:
            rep += "sets=%s" % self.sets
        if len(self.incs) > 0:
            rep += "incs=%s" % self.incs
        if len(self.reads) > 0:
            rep += "reads=%s" % self.reads
        if len(self.updates) > 0:
            rep += "updates=%s" % self.updates
        rep += ")"
        return rep

    def make_step(self, *args):
        """Raises an error (since this shouldn't be called)."""
        raise NotImplementedError()

    def init_signals(self, *args):
        """Raises an error (since this shouldn't be called)."""
        raise NotImplementedError()


@builder.Builder.register(Op)
class Builder(builder.OpBuilder):
    """
    Mock-up builder for `.Op`.
    """

    @staticmethod
    def mergeable(x, y):
        return True


class Probe(nengo.Probe):
    """
    Mock-up for `nengo.Probe`.
    """

    def __init__(self, target=None):
        # pylint: disable=super-init-not-called
        if target is not None:
            # bypass target validation
            nengo.Probe.target.data[self] = target

    @property
    def size_in(self):
        """Modified version of size_in that supports Signal targets."""

        return (
            self.target.size
            if isinstance(self.target, (Signal, nengo.builder.signal.Signal))
            else self.target.size_out
        )


class Simulator:
    """
    Mock-up for `nengo.Simulator`.
    """

    model = nengo.builder.Model()
    model.sig = defaultdict(lambda: defaultdict(Signal))


class TensorGraph(tensor_graph.TensorGraph):
    """
    Mock-up for `.tensor_graph.TensorGraph`.
    """

    def __init__(self, plan=None, dtype=None, minibatch_size=None):
        # pylint: disable=bad-super-call
        super(tensor_graph.TensorGraph, self).__init__(dtype=dtype)

        self.plan = plan
        self.minibatch_size = minibatch_size
        self.seed = 0
        self.model = nengo.builder.Model()
        self.invariant_inputs = {}

        self.signals = signals.SignalDict(self.dtype, self.minibatch_size)


def linear_net():
    """
    A simple network with an input, output, and no nonlinearity.
    """

    with nengo.Network() as net:
        a = nengo.Node([1])
        b = nengo.Node(size_in=1)
        nengo.Connection(a, b, synapse=None, transform=1)
        p = nengo.Probe(b)

    return net, a, p


def DeterministicLIF(**kwargs):
    """
    A LIF neuron with fixed initial voltages (useful in tests where we want neurons
    to always have the same output given the same parameters, but don't want to
    fix the seed).
    """

    if version.parse(nengo.__version__) <= version.parse("3.0.0"):
        return nengo.LIF(**kwargs)
    else:
        return nengo.LIF(initial_state={"voltage": nengo.dists.Choice([0])}, **kwargs)



# pylint: disable=missing-docstring

import contextlib
import logging
import os
import pickle
import sys
from collections import OrderedDict

import nengo
import numpy as np
import pytest
import tensorflow as tf
from nengo.dists import Uniform
from nengo.exceptions import (
    ReadonlyError,
    SimulationError,
    SimulatorClosed,
    ValidationError,
)
from packaging import version
from tensorflow.core.util import event_pb2
from tensorflow.python.eager import context

from nengo_dl import Layer, TensorNode, callbacks, configure_settings, dists, utils
from nengo_dl.compat import TFLogFilter, default_transform, eager_enabled
from nengo_dl.simulator import SimulationData
from nengo_dl.tests import dummies


def test_persistent_state(Simulator, seed):
    """Make sure that state is preserved between runs."""

    with nengo.Network(seed=seed) as net:
        configure_settings(dtype="float64")

        inp = nengo.Node(np.sin)
        ens = nengo.Ensemble(1000, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    with Simulator(net) as sim:
        sim.run_steps(100)
        data = sim.data[p]
        sim.reset()

        sim.run_steps(100)
        data2 = sim.data[p]
        sim.reset()

        for _ in range(100 // sim.unroll):
            sim.run_steps(sim.unroll)
        data3 = sim.data[p]

    assert np.allclose(data, data2)
    assert np.allclose(data2, data3)

    with net:
        configure_settings(use_loop=False)
    with Simulator(net) as sim:
        for _ in range(100 // sim.unroll):
            sim.run_steps(sim.unroll)
        data4 = sim.data[p]

    assert np.allclose(data3, data4)


def test_step_blocks(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        inp = nengo.Node(np.sin)
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    with Simulator(net, unroll_simulation=25) as sim1:
        sim1.run_steps(50)
    with Simulator(net, unroll_simulation=10) as sim2:
        sim2.run_steps(50)

    assert np.allclose(sim1.data[p], sim2.data[p])

    with pytest.warns(RuntimeWarning):
        with Simulator(net, unroll_simulation=5) as sim:
            sim.run_steps(2)


def test_minibatch(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        inp = [
            nengo.Node(output=[0.5]),
            nengo.Node(output=np.sin),
            nengo.Node(output=nengo.processes.WhiteSignal(5, 0.5, seed=seed)),
        ]

        ens = [
            nengo.Ensemble(10, 1, neuron_type=nengo.AdaptiveLIF()),
            nengo.Ensemble(10, 1, neuron_type=nengo.LIFRate()),
            nengo.Ensemble(10, 2, noise=nengo.processes.WhiteNoise(seed=seed)),
        ]

        nengo.Connection(inp[0], ens[0])
        nengo.Connection(inp[1], ens[1], synapse=None)
        nengo.Connection(inp[2], ens[2], synapse=nengo.Alpha(0.1), transform=[[1], [1]])
        conn = nengo.Connection(ens[0], ens[1], learning_rule_type=nengo.PES())
        nengo.Connection(inp[0], conn.learning_rule)

        ps = [nengo.Probe(e) for e in ens]

    with Simulator(net, minibatch_size=None) as sim:
        probe_data = [[] for _ in ps]
        for _ in range(5):
            sim.run_steps(100)

            for j, p in enumerate(ps):
                probe_data[j] += [sim.data[p]]

            sim.reset()

        probe_data = [np.stack(x, axis=0) for x in probe_data]

    with Simulator(net, minibatch_size=5) as sim:
        sim.run_steps(100)

    for i, p in enumerate(ps):
        assert np.allclose(sim.data[p], probe_data[i], atol=1e-6)


def test_input_feeds(Simulator):
    minibatch_size = 10

    with nengo.Network() as net:
        inp = nengo.Node([0, 0, 0])
        out = nengo.Node(size_in=3)
        nengo.Connection(inp, out, synapse=None)
        p = nengo.Probe(out)

        # check that sliced input nodes are fed properly as well
        inp2 = nengo.Node([0, 0, 0])
        out2 = nengo.Node(size_in=2)
        nengo.Connection(inp2[:2], out2, synapse=None)
        p2 = nengo.Probe(out2)

    with Simulator(net, minibatch_size=minibatch_size) as sim:
        val = np.random.randn(minibatch_size, 50, 3)
        sim.run_steps(50, data={inp: val, inp2: val})
        assert np.allclose(sim.data[p], val)
        assert np.allclose(sim.data[p2], val[..., :2])

        # error for wrong minibatch size
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim.run_steps(10, data={inp: np.zeros((minibatch_size * 2, 10, 3))})
        # error for wrong number of steps
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim.run_steps(10, data={inp: np.zeros((minibatch_size, 15, 3))})


@pytest.mark.parametrize(
    "neurons, use_loop", [(True, False), (False, False), (True, True)]
)
@pytest.mark.training
def test_train_ff(Simulator, neurons, use_loop, seed):
    minibatch_size = 4
    n_hidden = 20

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1])
        net.config[nengo.Ensemble].bias = nengo.dists.Choice([0])
        net.config[nengo.Connection].synapse = None

        # note: we have these weird input setup just so that we can test
        # training with two distinct inputs
        inp_a = nengo.Node([0])
        inp_b = nengo.Node([0])
        inp = nengo.Node(size_in=2)
        nengo.Connection(inp_a, inp[0], transform=1)
        nengo.Connection(inp_b, inp[1], transform=1)

        ens = nengo.Ensemble(
            n_hidden + 1, n_hidden, neuron_type=nengo.Sigmoid(tau_ref=1)
        )
        out = nengo.Ensemble(1, 1, neuron_type=nengo.Sigmoid(tau_ref=1))
        nengo.Connection(inp, ens.neurons if neurons else ens, transform=dists.Glorot())
        nengo.Connection(
            ens.neurons if neurons else ens, out.neurons, transform=dists.Glorot()
        )

        p = nengo.Probe(out.neurons)

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=1, seed=seed
    ) as sim:
        x = np.asarray([[[0.0, 0.0]], [[0.0, 1.0]], [[1.0, 0.0]], [[1.0, 1.0]]])
        y = np.asarray([[[0.1]], [[0.9]], [[0.9]], [[0.1]]])

        sim.compile(tf.optimizers.Adam(0.01), loss=tf.losses.mse)
        sim.fit({inp_a: x[..., [0]], inp_b: x[..., [1]]}, {p: y}, epochs=500, verbose=0)

        sim.step(data={inp_a: x[..., [0]], inp_b: x[..., [1]]})

        assert np.allclose(sim.data[p], y, atol=2e-3)


@pytest.mark.parametrize("truncation", (None, 5))
@pytest.mark.parametrize("use_loop", (True, False))
@pytest.mark.training
def test_train_recurrent(Simulator, truncation, use_loop, seed):
    batch_size = 100
    minibatch_size = 100
    n_hidden = 30
    n_steps = 10

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([0])
        ens = nengo.Ensemble(
            n_hidden,
            1,
            neuron_type=nengo.RectifiedLinear(),
            gain=np.ones(n_hidden),
            bias=np.zeros(n_hidden),
        )
        out = nengo.Node(size_in=1)

        nengo.Connection(inp, ens, synapse=None)
        nengo.Connection(ens.neurons, ens.neurons, transform=dists.He(), synapse=0)
        nengo.Connection(ens, out, synapse=None)

        p = nengo.Probe(out)

    kwargs = dict(unroll_simulation=truncation or n_steps) if use_loop else dict()

    with Simulator(net, minibatch_size=minibatch_size, seed=seed, **kwargs) as sim:
        x = np.outer(np.linspace(0, 1, batch_size), np.ones(n_steps))[:, :, None]
        y = np.outer(np.linspace(0, 1, batch_size), np.linspace(0, 1, n_steps))[
            :, :, None
        ]

        sim.compile(tf.optimizers.RMSprop(1e-3), loss=tf.losses.mse)

        if truncation:
            truncation_steps = n_steps // truncation

            # TODO: why doesn't the callback approach work?
            # class ResetCallback(tf.keras.callbacks.Callback):
            #     def on_train_batch_end(self, batch, logs=None):
            #         if batch % truncation_steps == truncation_steps - 1:
            #             sim.reset(
            #                 include_probes=False,
            #                 include_trainable=False,
            #                 include_processes=False,
            #             )
            #
            #     # def on_epoch_end(self, epoch, logs=None):
            #     #     sim.soft_reset()
            #
            # sim.fit(
            #     {inp: np.reshape(x, (-1, truncation, x.shape[2]))},
            #     {p: np.reshape(y, (-1, truncation, y.shape[2]))},
            #     epochs=200,
            #     shuffle=False,
            #     stateful=True,
            #     callbacks=[ResetCallback()],
            #     verbose=2,
            # )

            # TODO: why does this produce non-deterministic results?
            for _ in range(200):
                for j in range(truncation_steps):
                    sim.fit(
                        {inp: x[:, j * truncation : (j + 1) * truncation]},
                        {p: y[:, j * truncation : (j + 1) * truncation]},
                        epochs=1,
                        stateful=True,
                        verbose=0,
                    )

                # note: this works because each epoch is a single minibatch. otherwise
                # we would also have to reset after each minibatch (which is what
                # the callback approach is trying to do)
                sim.reset(
                    include_probes=False,
                    include_trainable=False,
                    include_processes=False,
                )
        else:
            sim.fit({inp: x}, {p: y}, epochs=200, verbose=0)

        loss = sim.evaluate({inp: x}, {p: y})["loss"]
        assert loss < (0.007 if truncation else 0.0025)


@pytest.mark.training
def test_recurrent_gradients(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(
            10, 1, gain=nengo.dists.Choice([0.5]), bias=nengo.dists.Choice([0])
        )
        nengo.Connection(a, b.neurons, transform=nengo.dists.Gaussian(0, 1))
        p = nengo.Probe(b.neurons)

    with Simulator(net) as sim:
        sim.check_gradients(inputs=[np.ones((1, sim.unroll * 2, 1)) * 0.1], outputs=[p])


@pytest.mark.parametrize("unroll", (1, 2))
@pytest.mark.training
def test_train_objective(Simulator, unroll, seed):
    minibatch_size = 1
    n_hidden = 20
    n_steps = 10

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([1])

        ens = nengo.Ensemble(n_hidden, 1, neuron_type=nengo.RectifiedLinear())
        nengo.Connection(inp, ens, synapse=0.01, transform=1)
        p = nengo.Probe(ens)

        ens2 = nengo.Ensemble(n_hidden, 1, neuron_type=nengo.RectifiedLinear())
        nengo.Connection(inp, ens2, synapse=0.01, transform=1)
        p2 = nengo.Probe(ens2)

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=unroll, seed=seed
    ) as sim:
        x = np.ones((minibatch_size, n_steps, 1))
        y = np.zeros((minibatch_size, n_steps, 1))
        z = np.zeros((minibatch_size, n_steps, 1)) + 0.1

        def obj(target, output):
            return tf.reduce_mean((output[:, -1] - 0.5 - target[:, -1]) ** 2)

        sim.compile(tf.optimizers.SGD(1e-2, momentum=0.9), loss={p: obj, p2: obj})
        sim.fit({inp: x}, {p: y, p2: z}, epochs=200, verbose=0)

        sim.check_gradients(outputs=[p, p2])

        sim.run_steps(n_steps, data={inp: x})

        assert np.allclose(sim.data[p][:, -1], y[:, -1] + 0.5, atol=1e-3)
        assert np.allclose(sim.data[p2][:, -1], z[:, -1] + 0.5, atol=1e-3)


@pytest.mark.training
def test_train_sparse(Simulator, seed):
    minibatch_size = 4
    n_hidden = 20

    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].gain = nengo.dists.Choice([1])
        net.config[nengo.Ensemble].bias = nengo.dists.Choice([0])
        net.config[nengo.Ensemble].neuron_type = nengo.RectifiedLinear()
        net.config[nengo.Connection].synapse = None

        inp = nengo.Node([0, 0, 0, 0, 0])
        ens = nengo.Ensemble(n_hidden, 1)
        out = nengo.Node(size_in=2)
        nengo.Connection(inp[[0, 2, 3]], ens.neurons, transform=dists.Glorot())
        nengo.Connection(ens.neurons, out, transform=dists.Glorot())

        p = nengo.Probe(out)

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=1, seed=seed
    ) as sim:
        x = np.asarray(
            [
                [[0.0, 0.0, 0.0, 0.0, 0.0]],
                [[0.0, 0.0, 1.0, 0.0, 0.0]],
                [[1.0, 0.0, 0.0, 0.0, 0.0]],
                [[1.0, 0.0, 1.0, 0.0, 0.0]],
            ]
        )
        y = np.asarray([[[0, 1]], [[1, 0]], [[1, 0]], [[0, 1]]])

        sim.compile(
            tf.optimizers.SGD(0.1, momentum=0.9, nesterov=True), loss=tf.losses.mse
        )
        sim.fit({inp: x}, {p: y}, epochs=500, verbose=0)

        sim.step(data={inp: x})

        assert np.allclose(sim.data[p], y, atol=1e-3)


@pytest.mark.training
def test_train_errors(Simulator):
    net, a, p = dummies.linear_net()

    n_steps = 20
    with Simulator(net) as sim:
        sim.compile(
            optimizer=tf.optimizers.SGD(0),
            loss=lambda y_true, y_pred: tf.losses.mse(y_true[:, -1], y_pred[:, -1]),
        )

        # error for mismatched batch size
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim.fit({a: np.ones((2, n_steps, 1))}, {p: np.ones((1, n_steps, 1))})

        # no error for mismatched n_steps
        sim.fit({a: np.ones((1, n_steps + 5, 1))}, {p: np.ones((1, n_steps, 1))})

    # error when calling train after closing
    with pytest.raises(SimulatorClosed, match="call fit"):
        sim.fit(n_steps=1)

    with Simulator(net, unroll_simulation=2) as sim:
        # error when data n_steps does not match unroll
        with pytest.raises(ValidationError, match="must be evenly divisible"):
            sim.fit({a: np.ones((1, 1, 1))}, {p: np.ones((1, 1, 1))})

        # error when n_steps does not evenly divide by truncation
        # TODO: support truncation
        # with pytest.raises(ValidationError):
        #     sim.train(
        #         {a: np.ones((1, 4, 1)), p: np.ones((1, 4, 1))}, None, truncation=3
        #     )


@pytest.mark.training
def test_train_no_data(Simulator):
    net, _, p = dummies.linear_net()

    with Simulator(net) as sim:

        sim.compile(
            tf.optimizers.SGD(0.1), loss={p: lambda y_true, y_pred: abs(2.0 - y_pred)}
        )
        sim.fit(n_steps=5, y=np.zeros((1, 5, 1)), epochs=10)

        sim.step()
        assert np.allclose(sim.data[p], 2)


def test_evaluate_errors(Simulator):
    with nengo.Network() as net:
        inp = nengo.Node([1])
        p = nengo.Probe(inp)

    n_steps = 20
    with Simulator(net, unroll_simulation=1) as sim:
        sim.compile(loss=lambda y_true, y_pred: 1.0)

        # check that valid inputs pass
        assert np.allclose(
            sim.evaluate(y={p: np.zeros((1, n_steps, 1))}, n_steps=n_steps)["loss"], 1
        )

        # error for incorrect n_steps (when explicitly specified)
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim.evaluate(
                {inp: np.ones((1, n_steps + 1, 1))},
                {p: np.ones((1, n_steps, 1))},
                n_steps=n_steps,
            )

        # no error for mismatched n_steps (between inputs and targets)
        sim.evaluate({inp: np.ones((1, n_steps, 1))}, {p: np.ones((1, n_steps + 1, 1))})

        # error for mismatched batch size (between inputs and targets)
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim.evaluate({inp: np.ones((1, n_steps, 1))}, {p: np.ones((2, n_steps, 1))})

        # must specify n_steps if no input data
        with pytest.raises(ValidationError, match="either input data or n_steps"):
            sim.evaluate(y={p: np.zeros((4, n_steps, 1))})

    # error when calling evaluate after close
    with pytest.raises(SimulatorClosed, match="call evaluate"):
        sim.evaluate(y={p: np.zeros((1, n_steps, 1))}, n_steps=n_steps)

    with Simulator(net, unroll_simulation=2) as sim:
        # error when data n_steps does not match unroll
        with pytest.raises(ValidationError, match="must be evenly divisible"):
            sim.evaluate({inp: np.ones((1, 1, 1))}, {p: np.ones((1, 1, 1))})


def test_generate_inputs(Simulator, seed):
    n_steps = 3
    minibatch_size = 2

    with nengo.Network() as net:
        proc = nengo.processes.WhiteNoise(seed=seed)
        inp = [
            nengo.Node([1]),
            nengo.Node(np.sin),
            nengo.Node(proc),
            nengo.Node([2]),
            nengo.Node(nengo.processes.WhiteNoise()),
        ]

        p = [nengo.Probe(x) for x in inp]

    with Simulator(
        net, minibatch_size=minibatch_size, unroll_simulation=n_steps
    ) as sim:
        feed = sim._generate_inputs(
            data={inp[0]: np.zeros((minibatch_size, n_steps, 1))}, n_steps=n_steps
        )

        assert len(sim.tensor_graph.invariant_inputs) == len(inp)
        assert len(feed) == len(inp) + 1

        sim.reset()
        sim.run_steps(n_steps, data={inp[0]: np.zeros((minibatch_size, n_steps, 1))})

        vals = [
            np.zeros((minibatch_size, n_steps, 1)),
            np.tile(np.sin(sim.trange())[None, :, None], (minibatch_size, 1, 1)),
            np.tile(proc.run_steps(3)[None, :, :], (minibatch_size, 1, 1)),
            np.ones((minibatch_size, n_steps, 1)) * 2,
        ]
        for i, x in enumerate(vals):
            assert np.allclose(feed["node_%d" % i if i > 0 else "node"], x)
            assert np.allclose(sim.data[p[i]], x)

        # check that unseeded process was different in each minibatch item
        assert not np.allclose(
            feed["node_%d" % (len(inp) - 1)][0], feed["node_%d" % (len(inp) - 1)][1]
        )

        with pytest.raises(SimulationError, match="automatically add n_steps"):
            sim._generate_inputs(data=range(5), n_steps=1)

        with pytest.raises(ValidationError, match="not a valid input"):
            sim._generate_inputs(data={p[0]: np.zeros((minibatch_size, n_steps, 1))})


@pytest.mark.parametrize("include_state", (True, False))
def test_save_load_params(Simulator, include_state, tmpdir):
    def get_network(seed):
        with nengo.Network(seed=seed) as net:
            configure_settings(simplifications=[])

            inp = nengo.Node([0])
            out = nengo.Node(size_in=1)
            ens = nengo.Ensemble(10, 1, neuron_type=dummies.DeterministicLIF())
            nengo.Connection(inp, ens)
            conn = nengo.Connection(ens, out)
            p = nengo.Probe(out)

            configure_settings(trainable=None)
            net.config[ens].trainable = False

        return net, ens, conn, p

    net0, ens0, conn0, p0 = get_network(0)

    with Simulator(net0) as sim_save:
        weights0, enc0, bias0 = sim_save.data.get_params(
            (conn0, "weights"), (ens0, "encoders"), (ens0, "bias")
        )

        sim_save.run_steps(10)

        sim_save.save_params(str(tmpdir), include_state=include_state)

        sim_save.run_steps(10)

    with pytest.raises(SimulatorClosed):
        sim_save.save_params(None)
    with pytest.raises(SimulatorClosed):
        sim_save.load_params(None)

    net1, ens1, conn1, p1 = get_network(1)

    with Simulator(net1) as sim_load:
        weights1, enc1, bias1 = sim_load.data.get_params(
            (conn1, "weights"), (ens1, "encoders"), (ens1, "bias")
        )
        assert not np.allclose(weights0, weights1)
        assert not np.allclose(enc0, enc1)
        assert not np.allclose(bias0, bias1)

        pre_model = sim_load.keras_model

        sim_load.load_params(str(tmpdir), include_state=include_state)

        weights2, enc2, bias2 = sim_load.data.get_params(
            (conn1, "weights"), (ens1, "encoders"), (ens1, "bias")
        )

        # check if params match
        assert np.allclose(weights0, weights2)
        assert np.allclose(enc0, enc2)
        assert np.allclose(bias0, bias2)

        # check if a new model was created or one was modified in-place
        assert sim_load.keras_model is pre_model

        # check if things still run correctly
        sim_load.run_steps(10)

        # check if simulation state resumed correctly
        if include_state:
            # state saved, so we should match the point at which that state was saved
            assert np.allclose(sim_load.data[p1], sim_save.data[p0][10:])
        else:
            # state not saved, but other seeded params are, so we should match the first
            # timesteps of `sim_save` (despite the networks not having the same seeds)
            assert np.allclose(sim_load.data[p1], sim_save.data[p0][:10])

    with Simulator(nengo.Network()) as sim:
        with pytest.raises(SimulationError, match="!= number of variables"):
            sim.load_params(str(tmpdir))


def test_save_load_params_deprecation(Simulator, tmpdir):
    with nengo.Network() as net:
        a = nengo.Node([1])
        p = nengo.Probe(a, synapse=0.1)

    with Simulator(net) as sim0:
        sim0.run_steps(5)

        with pytest.warns(
            DeprecationWarning, match="include_non_trainable is deprecated"
        ):
            sim0.save_params(str(tmpdir.join("tmp")), include_non_trainable=True)

        sim0.run_steps(5)

    with Simulator(net) as sim1:
        with pytest.warns(
            DeprecationWarning, match="include_non_trainable is deprecated"
        ):
            sim1.load_params(str(tmpdir.join("tmp")), include_non_trainable=True)

        sim1.run_steps(5)

    assert np.allclose(sim0.data[p][-5:], sim1.data[p])


def test_model_passing(Simulator, seed):
    # make sure that passing a built model to the Simulator works properly

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([1])
        ens = nengo.Ensemble(20, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    model = nengo.builder.Model()
    model.build(net)

    ops = model.operators.copy()

    with nengo.Simulator(None, model=model, optimize=False) as sim:
        sim.run_steps(10)

    assert ops == model.operators
    canonical = sim.data[p]

    with Simulator(None, model=model) as sim:
        sim.run_steps(10)

    assert ops == model.operators
    assert np.allclose(sim.data[p], canonical)

    # make sure that passing the same model to Simulator twice works
    with Simulator(None, model=model) as sim:
        sim.run_steps(10)
        assert ops == model.operators

    assert ops == model.operators
    assert np.allclose(sim.data[p], canonical)

    # make sure that passing that model back to the reference simulator works
    with nengo.Simulator(None, model=model, optimize=False) as sim:
        sim.run_steps(10)

    assert ops == model.operators
    assert np.allclose(sim.data[p], canonical)


@pytest.mark.parametrize("device", ["/cpu:0", "/gpu:0", None])
def test_devices(Simulator, device, seed, caplog):
    if device == "/gpu:0" and not utils.tf_gpu_installed:
        pytest.skip("This test requires GPU support")

    caplog.set_level(logging.INFO)

    with nengo.Network(seed=seed) as net:
        inp = nengo.Node([0])
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens)

    with nengo.Simulator(net) as sim:
        sim.run_steps(50)
        canonical = sim.data[p]

    with Simulator(net, device=device) as sim:
        sim.run_steps(50)
        assert np.allclose(sim.data[p], canonical)

        if device == "/cpu:0":
            assert "Running on CPU" in caplog.text
        elif device == "/gpu:0":
            assert "Running on GPU" in caplog.text
        elif utils.tf_gpu_installed:
            assert "Running on CPU/GPU" in caplog.text
        else:
            # device is None but gpu not installed
            assert "Running on CPU" in caplog.text


def test_side_effects(Simulator):
    class MyFunc:
        x = 0

        def __call__(self, t):
            self.x += 1

    func = MyFunc()

    with nengo.Network() as net:
        nengo.Node(output=func)

    with Simulator(net, unroll_simulation=1) as sim:
        sim.run_steps(10)

    assert func.x == 11


@pytest.mark.training
def test_tensorboard(Simulator, tmpdir):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1, neuron_type=nengo.LIFRate())
        c = nengo.Connection(a, b, transform=1)
        c0 = nengo.Connection(a, b)
        p = nengo.Probe(b)
        p2 = nengo.Probe(c)

    # check that training summaries are output properly
    n_epochs = 3
    with Simulator(net) as sim:

        log_dir = str(tmpdir.join("a_run"))

        sim.compile(
            tf.optimizers.SGD(0.0),
            loss={p: lambda y_true, y_pred: y_pred, p2: tf.losses.mse},
        )

        sim.fit(
            {a: np.zeros((1, 10, 1))},
            {p: np.zeros((1, 10, 1)), p2: np.zeros((1, 10, 1))},
            epochs=n_epochs,
            callbacks=[
                tf.keras.callbacks.TensorBoard(log_dir=log_dir, profile_batch=0),
                callbacks.NengoSummaries(
                    log_dir=os.path.join(log_dir, "nengo"),
                    sim=sim,
                    objects=[b, b.neurons, c],
                ),
            ],
        )

    # look up name of event file
    event_dir = (
        os.path.join(log_dir, "train")
        if version.parse(tf.__version__) < version.parse("2.3.0rc0") or eager_enabled()
        else log_dir
    )
    event_file = [
        x
        for x in os.listdir(event_dir)
        if x.startswith("events.out.tfevents") and not x.endswith(".profile-empty")
    ]
    assert len(event_file) == 1
    event_file = os.path.join(event_dir, event_file[0])
    assert os.path.exists(event_file)

    summaries = ["epoch_loss", "epoch_probe_loss", "epoch_probe_1_loss"]
    # metadata stuff in event file
    meta_steps = (
        2
        if version.parse(tf.__version__) < version.parse("2.3.0rc0") or eager_enabled()
        else 3
    )
    with contextlib.suppress() if eager_enabled() else context.eager_mode():
        for i, record in enumerate(tf.data.TFRecordDataset(event_file)):
            event = event_pb2.Event.FromString(record.numpy())

            if i >= meta_steps:
                curr_step = (i - meta_steps) // len(summaries)
                assert event.step == curr_step

                # assert event.summary.value[0].tag
                # == summaries[(i - 2) % len(summaries)]
                # log order non-deterministic in python 3.5 so we use this less
                # stringent check
                assert event.summary.value[0].tag in summaries

    assert i == len(summaries) * n_epochs + (  # pylint: disable=undefined-loop-variable
        meta_steps - 1
    )

    # look up name of event file
    event_file = [
        x for x in os.listdir(os.path.join(log_dir, "nengo")) if x.endswith(".v2")
    ]
    assert len(event_file) == 1
    event_file = os.path.join(log_dir, "nengo", event_file[0])
    assert os.path.exists(event_file)

    summaries = [
        "Ensemble_None_encoders",
        "Ensemble.neurons_None_bias",
        "Connection_None_weights",
    ]
    with contextlib.suppress() if eager_enabled() else context.eager_mode():
        for i, record in enumerate(tf.data.TFRecordDataset(event_file)):
            event = event_pb2.Event.FromString(record.numpy())

            if i < 1:
                # metadata stuff
                continue

            curr_step = (i - 1) // len(summaries)
            assert event.step == curr_step
            assert event.summary.value[0].tag == summaries[(i - 1) % len(summaries)]

    assert i == len(summaries) * n_epochs  # pylint: disable=undefined-loop-variable

    # check for error on invalid object
    with pytest.raises(ValidationError, match="Unknown summary object"):
        callbacks.NengoSummaries(log_dir=log_dir + "/nengo", sim=sim, objects=[a])

    if version.parse(nengo.__version__) >= version.parse("3.1.0"):
        with pytest.raises(ValidationError, match="does not have any weights"):
            callbacks.NengoSummaries(log_dir=log_dir + "/nengo", sim=sim, objects=[c0])


@pytest.mark.parametrize("mode", ("predict", "train"))
@pytest.mark.training
def test_profile(Simulator, mode, tmpdir, pytestconfig):
    if (
        pytestconfig.getoption("--graph-mode")
        and version.parse(tf.__version__) >= version.parse("2.4.0rc0")
        and mode == "predict"
    ):
        pytest.skip(
            "TensorFlow bug, see https://github.com/tensorflow/tensorflow/issues/44563"
        )

    net, a, p = dummies.linear_net()

    with Simulator(net) as sim:
        # note: TensorFlow bug if using profile_batch=1, see
        # https://github.com/tensorflow/tensorflow/issues/37543
        callback = callbacks.TensorBoard(
            log_dir=str(tmpdir.join("profile")), profile_batch=2
        )

        if mode == "predict":
            sim.predict(np.zeros((2, 5, 1)), callbacks=[callback])
        else:
            sim.compile(tf.optimizers.SGD(1), loss=tf.losses.mse)
            sim.fit(
                {a: np.zeros((2, 5, 1))}, {p: np.zeros((2, 5, 1))}, callbacks=[callback]
            )

        assert os.path.exists(
            str(
                tmpdir.join("profile", "train")
                if version.parse(tf.__version__) < version.parse("2.3.0rc0")
                else tmpdir.join("profile")
            )
        )


def test_dt_readonly(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        nengo.Probe(a)

    with Simulator(net) as sim:
        with pytest.raises(ReadonlyError):
            sim.dt = 1


def test_probe_data():
    sim = dummies.Simulator()
    a = dummies.Probe(add_to_container=False)
    b = dummies.Probe(add_to_container=False)
    sim.model.params = OrderedDict(
        {
            a: [np.zeros((5, 1, 3)), np.ones((5, 1, 3))],
            b: [np.ones((1, 1, 3)), np.zeros((1, 1, 3))],
        }
    )
    sim.model.probes = (a, b)
    data = SimulationData(sim, True)
    assert data[a].shape == (5, 2, 3)
    assert np.all(data[a][:, 0] == 0)
    assert np.all(data[a][:, 1] == 1)

    data.minibatched = False
    assert data[b].shape == (2, 3)
    assert np.all(data[b][1] == 0)
    assert np.all(data[b][0] == 1)


@pytest.mark.parametrize(
    "pre_val", [1, lambda t: 1, nengo.processes.Piecewise({0: [1]})]
)
@pytest.mark.parametrize(
    "post_val", [2, lambda t: 2, nengo.processes.Piecewise({0: [2]})]
)
def test_node_output_change(Simulator, pre_val, post_val, seed):
    with nengo.Network(seed=seed) as net:
        inp = nengo.Node(pre_val)
        p = nengo.Probe(inp)

    with Simulator(net, unroll_simulation=1) as sim:
        sim.step()
        inp.output = post_val
        sim.step()

        assert np.allclose(sim.data[p], 1.0)

        sim.reset()
        sim.step()

        assert np.allclose(sim.data[p], 1.0)


@pytest.mark.training
def test_check_gradients_error(Simulator):
    # check_gradients detects nans in gradient
    with nengo.Network() as net:
        x = nengo.Node([0])
        y = Layer(lambda x: 1 / x)(x)
        nengo.Probe(y)

    with Simulator(net) as sim:
        with pytest.raises(SimulationError, match="NaNs detected"):
            sim.check_gradients()

    # check_gradients detects errors in gradient (in this case caused by the
    # fact that nengo.Triangle doesn't have a TensorFlow implementation)
    with nengo.Network() as net:
        x = nengo.Node([0])
        nengo.Probe(x, synapse=nengo.Triangle(0.1))

    with Simulator(net) as sim:
        with pytest.raises(SimulationError, match="Gradient check failed"):
            sim.check_gradients()


def test_check_data(Simulator):
    with nengo.Network() as net:
        inpa = nengo.Node([0, 0], label="inpa")
        inpb = nengo.Node([0], label="inpb")
        nengo.Node(size_in=2, label="n")
        nengo.Probe(inpa, label="pa")
        nengo.Probe(inpb, label="pb")
        nengo.Ensemble(10, 1)

    with nengo.Network():
        inp2 = nengo.Node([0, 0], label="inp2")
        nengo.Probe(inp2, label="p2")

    with Simulator(net, minibatch_size=3, unroll_simulation=1) as sim:
        zeros1 = np.zeros((3, 1, 1))
        zeros2 = np.zeros((3, 1, 2))
        n_steps = np.ones((3, 1))

        # make sure that valid inputs pass
        sim._check_data({"inpa": zeros2, "inpb": zeros1, "n_steps": n_steps})
        sim._check_data({"pa": zeros2, "pb": zeros1}, nodes=False)

        # invalid input objects
        with pytest.raises(ValidationError, match="not a valid node name"):
            sim._check_data({"inp2": zeros2, "n_steps": n_steps})
        with pytest.raises(ValidationError, match="not a valid node name"):
            sim._check_data({"n": zeros2, "n_steps": n_steps})

        # invalid target object
        with pytest.raises(ValidationError, match="not a valid probe name"):
            sim._check_data({"p2": zeros2}, nodes=False)

        # mismatched input data
        with pytest.raises(ValidationError, match="different batch size"):
            sim._check_data(
                {
                    "inpa": zeros2,
                    "inpb": np.zeros((sim.minibatch_size * 2, 1, 1)),
                    "n_steps": n_steps,
                }
            )
        with pytest.raises(ValidationError, match="different number of timesteps"):
            sim._check_data(
                {"inpa": zeros2, "inpb": np.zeros((3, 2, 1)), "n_steps": n_steps}
            )

        # mismatched target data
        with pytest.raises(ValidationError, match="different batch size"):
            sim._check_data(
                {"pa": zeros2, "pb": np.zeros((sim.minibatch_size * 2, 1, 1))},
                nodes=False,
            )
        # no error for different n_steps
        sim._check_data({"pa": zeros2, "pb": np.zeros((3, 2, 1))}, nodes=False)

        # data that doesn't match explicit validation value
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim._check_data({"inpa": zeros2, "n_steps": np.ones((2, 1))}, batch_size=2)
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim._check_data(
                {"inpa": zeros2, "n_steps": np.ones_like(n_steps) * 2}, n_steps=2
            )
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim._check_data({"pa": zeros2}, batch_size=2, nodes=False)
        with pytest.raises(ValidationError, match="does not match expected size"):
            sim._check_data({"pa": zeros2}, n_steps=2, nodes=False)

        # data with batch size < minibatch_size
        with pytest.raises(ValidationError, match=r"Batch size of data \(1\)"):
            sim._check_data({"inpa": zeros2[[0]], "n_steps": n_steps})
        with pytest.raises(ValidationError, match=r"Batch size of data \(1\)"):
            sim._check_data({"pa": zeros2[[0]]}, nodes=False)

        # data with incorrect rank
        with pytest.raises(ValidationError, match="should have rank 3"):
            sim._check_data({"inpa": zeros2[0], "n_steps": n_steps})
        with pytest.raises(ValidationError, match="should have rank 3"):
            sim._check_data({"pa": zeros2[0]}, nodes=False)

        # no n_steps
        with pytest.raises(ValidationError, match="Must specify 'n_steps'"):
            sim._check_data({})

        # n_steps wrong shape
        with pytest.raises(ValidationError, match="wrong shape"):
            sim._check_data({"n_steps": np.asarray(10)})

        # non-constant n_steps
        with pytest.raises(ValidationError, match="have the same value"):
            sim._check_data({"n_steps": np.asarray([[0], [1], [2]])})

        # n_steps mismatch
        with pytest.raises(ValidationError, match="does not match"):
            sim._check_data({"n_steps": np.asarray([[10], [10], [10]])}, n_steps=5)

    with net:
        configure_settings(use_loop=False)

    with Simulator(net, unroll_simulation=1) as sim:
        # n_steps doesn't match loop length
        with pytest.raises(ValidationError, match="use_loop=False"):
            sim._check_data({"n_steps": np.asarray([[2]])}, n_steps=2)


def test_matching_node_out(Simulator):
    # make sure that nodes with identical outputs are handled correctly (the
    # outputs will have the same hash, so that can cause some errors)

    with nengo.Network() as net:
        a = nengo.Node(output=nengo.processes.WhiteSignal(1, 10), size_out=2)
        b = nengo.Node(output=nengo.processes.WhiteSignal(1, 10), size_out=2)
        c = nengo.Node(output=nengo.processes.WhiteSignal(1, 10), size_out=3)

        p_a = nengo.Probe(a)
        p_b = nengo.Probe(b)
        p_c = nengo.Probe(c)

    with Simulator(net) as sim:
        sim.run_steps(10)

        assert sim.data[p_a].shape == (10, 2)
        assert sim.data[p_b].shape == (10, 2)
        assert not np.allclose(sim.data[p_a], sim.data[p_b])
        assert sim.data[p_c].shape == (10, 3)


def test_probe_no_data(Simulator):
    with nengo.Network() as net:
        u = nengo.Node([0])
        p = nengo.Probe(u)

    with Simulator(net) as sim:
        pass

    assert sim.data[p] == []


@pytest.mark.training
def test_train_state_save(Simulator):
    with nengo.Network() as net:
        u = nengo.Node([1])
        o = nengo.Node(size_in=1)
        nengo.Connection(u, o)
        p = nengo.Probe(o, synapse=0.1)

    with Simulator(net) as sim:
        sim.run_steps(20)

    with Simulator(net) as sim2:
        sim2.run_steps(10)

        sim2.compile(tf.optimizers.SGD(0), loss=tf.losses.mse)
        sim2.fit({u: np.ones((4, 10, 1))}, {p: np.ones((4, 10, 1))})

        sim2.evaluate({u: np.ones((4, 10, 1))}, {p: np.ones((4, 10, 1))})

        sim2.run_steps(10)

    assert np.allclose(sim.data[p], sim2.data[p])


def test_simulation_data(Simulator, seed):
    rng = np.random.RandomState(seed)
    N = 17
    d = 2

    gain = rng.uniform(low=1, high=2, size=N)
    bias = rng.uniform(low=0.2, high=1, size=N)
    enc = rng.uniform(-1, 1, size=(N, d))
    enc /= np.linalg.norm(enc, axis=1)[:, None]

    with nengo.Network() as net:
        u = nengo.Node([0] * d)
        a = nengo.Ensemble(N, d, gain=gain, bias=bias, encoders=enc, radius=3)
        b = nengo.Ensemble(
            N,
            d,
            gain=gain,
            bias=bias,
            encoders=enc * 2,
            radius=3,
            normalize_encoders=False,
        )
        c = nengo.Ensemble(N, d, neuron_type=nengo.Direct())
        conn0 = nengo.Connection(u, a)
        conn = nengo.Connection(
            a.neurons,
            b,
            transform=rng.uniform(-1, 1, size=(d, N)),
            learning_rule_type=nengo.PES(),
        )

        p = nengo.Probe(b)

    with Simulator(net) as sim:
        # check dict functions
        assert u in sim.data
        assert a in sim.data
        assert b in sim.data
        assert c in sim.data
        assert conn in sim.data
        assert conn0 in sim.data
        assert p in sim.data
        assert net in sim.data
        assert conn.learning_rule in sim.data
        assert len(sim.data) == 10  # 9 objects + probe connection
        for k, k2 in zip(sim.data, sim.data.keys()):
            assert k is k2

        # check gain/bias
        assert np.allclose(gain, sim.data[a].gain)
        assert np.allclose(bias, sim.data[a].bias)

        # check max_rates/intercepts
        max_rates, intercepts = a.neuron_type.max_rates_intercepts(gain, bias)
        assert np.allclose(max_rates, sim.data[a].max_rates)
        assert np.allclose(intercepts, sim.data[a].intercepts)

        # check encoders/scaled_encoders
        assert np.allclose(enc, sim.data[a].encoders)
        assert np.allclose(enc * gain[:, None] / a.radius, sim.data[a].scaled_encoders)

        # make sure that the inferences still work with non-normalized encoders
        assert np.allclose(enc * 2, sim.data[b].encoders)
        assert np.allclose(gain, sim.data[b].gain)
        assert np.allclose(bias, sim.data[b].bias)
        assert np.allclose(
            enc * 2 * gain[:, None] / b.radius, sim.data[b].scaled_encoders
        )

        # check connection weights
        transform = conn.transform.init
        assert np.allclose(transform, sim.data[conn].weights)

        # check that batch dimension eliminated
        assert sim.data[conn].weights.shape == (d, N)

        # check that values can be updated live
        sig = sim.model.sig[a]["encoders"]
        tensor_sig = sim.tensor_graph.signals[sig]

        base = sim.tensor_graph.base_params[tensor_sig.key]
        tf.keras.backend.set_value(
            base, np.ones(base.shape, dtype=base.dtype.as_numpy_dtype())
        )

        assert np.allclose(sim.data[a].scaled_encoders, 1)
        assert np.allclose(sim.data[a].gain, np.sqrt(2) * a.radius)
        assert np.allclose(sim.data[a].encoders, 1 / np.sqrt(2))

        # check that direct mode ensemble parameters are handled correctly
        assert np.allclose(sim.data[c].encoders, np.eye(d))
        assert sim.data[c].gain is None
        assert sim.data[c].bias is None

    # check minibatch dimension moved to front
    with Simulator(net, minibatch_size=3) as sim:
        assert sim.data[conn].weights.shape == (3, d, N)

    # reverts back to init after simulator close, and warns
    with pytest.warns(UserWarning):
        assert np.allclose(sim.data[a].encoders, enc)

    with pytest.raises(ValidationError):
        _ = sim.data[nengo.Ensemble(10, 1, add_to_container=False)]


@pytest.mark.training
def test_learning_rate_schedule(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1, neuron_type=nengo.LIFRate())
        nengo.Connection(a, b)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        vals = [1.0, 0.1, 0.001]

        def schedule(epoch, lr):
            if epoch >= 9:
                return vals[2]
            if epoch >= 4:
                return vals[1]
            return vals[0]

        sim.compile(tf.optimizers.SGD(vals[0]), loss=tf.losses.mse)

        for i in range(3):
            assert np.allclose(
                tf.keras.backend.get_value(sim.keras_model.optimizer.lr), vals[i]
            )
            sim.fit(
                {a: np.zeros((1, 10, 1))},
                {p: np.zeros((1, 10, 1))},
                epochs=(i + 1) * 5,
                initial_epoch=i * 5,
                callbacks=[tf.keras.callbacks.LearningRateScheduler(schedule)],
            )


@pytest.mark.training
def test_multiple_objective(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        net.config[nengo.Connection].synapse = None

        a = nengo.Node([0])

        # note: b is configured this way so that the output, and therefore
        # loss, will be equal to the input, so we can control it easily
        b = nengo.Ensemble(
            100,
            1,
            neuron_type=nengo.RectifiedLinear(),
            gain=nengo.dists.Choice([1]),
            bias=nengo.dists.Choice([0]),
            encoders=nengo.dists.Choice([[1]]),
        )

        c = nengo.Ensemble(10, 1)

        nengo.Connection(a, b)
        nengo.Connection(a, c)

        p_b = nengo.Probe(b)
        p_c = nengo.Probe(c)

    with Simulator(net, unroll_simulation=1) as sim:
        data = (
            {a: np.ones((10, 1, 1))},
            {p_b: np.zeros((10, 1, 1)), p_c: np.zeros((10, 1, 1))},
        )
        objective = {
            p_b: lambda y_true, y_pred: y_pred,
            p_c: lambda y_true, y_pred: y_pred * 0,
        }

        sim.compile(tf.optimizers.SGD(1.0), loss=objective)

        loss = sim.evaluate(*data)
        assert np.allclose(loss["loss"], 1, atol=1e-3)

        b_bias = np.copy(sim.data[b].bias)
        c_bias = sim.data[c].bias
        sim.fit(*data, epochs=10)
        assert np.allclose(sim.data[c].bias, c_bias)
        assert not np.allclose(sim.data[b].bias, b_bias)


def test_get_nengo_params(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        net.config[nengo.Ensemble].neuron_type = dummies.DeterministicLIF()

        a = nengo.Ensemble(12, 3, label="a")
        b = nengo.Ensemble(10, 4, label="b", radius=2)
        n = nengo.Node([1])
        c = nengo.Connection(a.neurons[:5], b[:2], transform=Uniform(-1, 1), label="c")
        d = nengo.Connection(
            a,
            b.neurons,
            function=lambda x: np.ones(5),
            transform=Uniform(-1, 1),
            label="d",
        )
        e = nengo.Connection(n, b, transform=Uniform(-1, 1), label="e")
        f = nengo.Ensemble(5, 1, label="a")
        g = nengo.Ensemble(11, 1, neuron_type=nengo.Direct(), label="g")
        h = nengo.Connection(
            a.neurons,
            b,
            transform=nengo.Convolution(1, (2, 2, 3), padding="same"),
            label="h",
        )
        p = nengo.Probe(b.neurons)

    with Simulator(net, seed=seed) as sim:
        # check that we get an error for non-ensemble/connection objects
        with pytest.raises(ValueError):
            sim.get_nengo_params(n)

        # check that we get an error for duplicate labels
        with pytest.raises(ValueError):
            sim.get_nengo_params([a, f], as_dict=True)

        # check that we get an error for direct ensembles
        with pytest.raises(ValueError):
            sim.get_nengo_params(g)

        # check that single objects are returned as single dicts
        params = sim.get_nengo_params(d)
        assert params["transform"] is default_transform

        fetches = [a.neurons, b, c, d, e, h]

        params = sim.get_nengo_params(fetches, as_dict=True)
        sim.run_steps(100)

    with nengo.Network(seed=seed + 1) as net:
        net.config[nengo.Ensemble].neuron_type = dummies.DeterministicLIF()

        a2 = nengo.Ensemble(12, 3, **params["a"])
        b2 = nengo.Ensemble(10, 4, radius=2, **params["b"])
        n2 = nengo.Node([1])
        nengo.Connection(a2.neurons[:5], b2[:2], **params["c"])
        nengo.Connection(a2, b2.neurons, **params["d"])
        nengo.Connection(n2, b2, **params["e"])
        nengo.Connection(a2.neurons, b2, **params["h"])
        p2 = nengo.Probe(b2.neurons)

    with Simulator(net, seed=seed) as sim2:
        sim2.run_steps(100)

        assert np.allclose(sim.data[p], sim2.data[p2])


@pytest.mark.parametrize("use_scipy", (True, False))
def test_get_nengo_params_sparse(Simulator, use_scipy, rng):
    if use_scipy:
        scipy_sparse = pytest.importorskip("scipy.sparse")

    with nengo.Network() as net:
        node0 = nengo.Node([1] * 10)
        node1 = nengo.Node(size_in=10)
        data = rng.uniform(-1, 1, size=10)
        if use_scipy:
            conn = nengo.Connection(
                node0,
                node1,
                synapse=None,
                transform=nengo.Sparse(
                    (10, 10),
                    init=scipy_sparse.coo_matrix(
                        (data, (np.arange(10), np.arange(10)))
                    ),
                ),
            )
        else:
            conn = nengo.Connection(
                node0,
                node1,
                synapse=None,
                transform=nengo.Sparse(
                    (10, 10), indices=[[i] * 2 for i in range(10)], init=data
                ),
            )

        p0 = nengo.Probe(node1)

    with Simulator(net) as sim0:
        params = sim0.get_nengo_params(conn)
        assert np.allclose(params["transform"].init.data, data)

        sim0.run_steps(10)

    with nengo.Network() as net:
        node0 = nengo.Node([1] * 10)
        node1 = nengo.Node(size_in=10)
        nengo.Connection(node0, node1, synapse=None, **params)
        p1 = nengo.Probe(node1)

    with Simulator(net) as sim1:
        sim1.run_steps(10)

    assert np.allclose(sim0.data[p0], sim1.data[p1])


@pytest.mark.parametrize("mixed", (False, True))
@pytest.mark.training
def test_direct_grads(Simulator, mixed):
    net, a, p = dummies.linear_net()

    if mixed:
        with net:
            c = nengo.Node(size_in=1)
            nengo.Connection(a, c, synapse=None, transform=1)
            p2 = nengo.Probe(c)

    with Simulator(net, minibatch_size=1) as sim:
        n_steps = 10

        def direct(y_true, y_pred):
            return tf.reduce_sum(y_true * y_pred)

        obj = {p: direct}
        if mixed:
            obj[p2] = tf.losses.mse

        sim.compile(tf.optimizers.SGD(0.45), loss=obj)

        for _ in range(10):
            sim.run_steps(n_steps)

            data = {p: (2.0 / n_steps) * (sim.data[p] - 2)}

            if mixed:
                data[p2] = np.ones((1, n_steps, 1)) * 2
                assert np.allclose(sim.data[p], sim.data[p2])

            sim.fit({a: np.ones((1, n_steps, 1))}, data)
            sim.reset(include_trainable=False, include_processes=False)

        sim.run_steps(n_steps)
        assert np.allclose(sim.data[p], 2)
        if mixed:
            assert np.allclose(sim.data[p2], 2)


@pytest.mark.training
def test_non_differentiable(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Node(lambda t, x: x, size_in=1)
        c = nengo.Connection(a, b, transform=1)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        w0 = sim.data[c].weights
        sim.compile(tf.optimizers.SGD(100), loss=tf.losses.mse)
        sim.fit({a: np.ones((1, 10, 1))}, {p: np.ones((1, 10, 1))})

        # TODO: find another way to detect non-differentiable elements in graph
        # note: the challenge is that our stateful ops tend to mask the
        # backpropagating None gradients, because the overwritten parts of
        # the variable end up with a gradient of zero (regardless of what the
        # output gradient is)
        assert np.allclose(sim.data[c].weights, w0)


@pytest.mark.parametrize(
    "net",
    (
        nengo.networks.BasalGanglia(4),
        nengo.networks.CircularConvolution(32, 4),
        nengo.networks.Oscillator(0.01, 10, 100),
        nengo.networks.InputGatedMemory(32, 4),
    ),
)
def test_freeze_network(Simulator, net):
    with Simulator(net) as sim:
        sim.run(0.1)
        sim.freeze_params(net)

    with nengo.Simulator(net) as sim2:
        sim2.run(0.1)

    for p in net.all_probes:
        assert np.allclose(sim.data[p], sim2.data[p])


def test_freeze_obj(Simulator, seed):
    with nengo.Network(seed=seed) as net:
        a = nengo.Ensemble(10, 1)
        b = nengo.Node(size_in=1)
        c = nengo.Connection(a, b)
        p = nengo.Probe(b)

    with nengo.Network(seed=seed + 1):
        d = nengo.Ensemble(10, 1)

    with Simulator(net) as sim:
        # check for error with object outside network
        with pytest.raises(ValueError):
            sim.freeze_params(d)

        # check for error with wrong object type
        with pytest.raises(TypeError):
            sim.freeze_params(p)

        sim.freeze_params([a, c])
        sim.run_steps(10)

    with nengo.Simulator(net) as sim2:
        sim2.run_steps(10)

    assert np.allclose(sim.data[p], sim2.data[p])

    # check that we get an error if the simulator is closed
    with pytest.raises(SimulatorClosed):
        sim.freeze_params(net)


@pytest.mark.training
def test_freeze_train(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([1])
        b = nengo.Node(size_in=1)
        nengo.Connection(a, b, transform=1, synapse=None)
        p = nengo.Probe(b)

    with Simulator(net, unroll_simulation=1) as sim:
        sim.step()
        assert np.allclose(sim.data[p], 1)

        sim.compile(tf.optimizers.SGD(0.5), loss=tf.losses.mse)
        sim.fit({a: np.ones((1, 1, 1))}, {p: np.ones((1, 1, 1)) * 2}, epochs=10)

        sim.step()
        assert np.allclose(sim.data[p][-1], 2)

        sim.freeze_params(net)

    with nengo.Simulator(net) as sim:
        sim.step()
        assert np.allclose(sim.data[p], 2)


@pytest.mark.parametrize("neuron_type", (nengo.SpikingRectifiedLinear(), nengo.LIF()))
def test_inference_only(Simulator, neuron_type, seed):
    with nengo.Network(seed=seed) as net:
        configure_settings(inference_only=False)

        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1, neuron_type=neuron_type)
        c = nengo.Connection(a, b, synapse=None, transform=1)
        p = nengo.Probe(b)

    with Simulator(net) as sim:
        sim.run(0.1)

        assert sim.model.sig[c]["weights"].trainable

    with net:
        configure_settings(inference_only=True)

    with Simulator(net) as sim2:
        sim2.run(0.1)

        # check that inference-only mode produces the same output
        assert np.allclose(sim.data[p], sim2.data[p])

        # check that parameters aren't trainable
        assert not sim2.model.sig[c]["weights"].trainable

        # validation checks (can't do train/gradients in inference-only mode)
        with pytest.raises(SimulationError, match="inference_only=True"):
            sim2.fit(n_steps=10)

        with pytest.raises(SimulationError, match="inference_only=True"):
            sim2.check_gradients()


@pytest.mark.training
def test_synapse_warning(Simulator):
    with nengo.Network() as net:
        a = nengo.Node([0])
        b = nengo.Ensemble(10, 1)
        c = nengo.Connection(a, b, synapse=1)
        p = nengo.Probe(b)
        p2 = nengo.Probe(b)

    def does_warn(n_steps=1, as_dict=True):
        with Simulator(net, unroll_simulation=1) as sim:
            sim.compile(tf.optimizers.SGD(0), loss={p: tf.losses.mse})
            with pytest.warns(None) as rec:
                if as_dict:
                    sim.fit(
                        {a: np.zeros((1, n_steps, 1))}, {p: np.zeros((1, n_steps, 1))}
                    )
                else:
                    sim.fit(n_steps=n_steps)
        return any("contains synaptic filters" in str(w.message) for w in rec)

    # warning from connection
    assert does_warn()

    # warning from probe
    c.synapse = None
    p.synapse = 1
    assert does_warn()

    # no warning for >1 step training
    assert not does_warn(n_steps=2)

    # no warning from non-target probe
    p.synapse = None
    p2.synapse = 1
    assert not does_warn()

    # warning when explicitly specifying n_steps
    c.synapse = 1
    assert does_warn(as_dict=True)


@pytest.mark.training
def test_concat_hang(Simulator, pytestconfig):
    with nengo.Network() as net:
        a = nengo.Node([0])
        x = nengo.Node(size_in=1)
        nengo.Connection(a, x)
        p = nengo.Probe(x)

    with Simulator(net) as sim:
        sim.compile(tf.optimizers.SGD(1), loss=tf.losses.mse)
        sim.fit({a: np.zeros((1, 5, 1))}, {p: np.zeros((1, 5, 1))})


def test_tf_seed(Simulator, seed):
    with nengo.Network() as net:
        a = TensorNode(lambda t: tf.random.uniform((1, 1), dtype=t.dtype))
        p = nengo.Probe(a)

    with Simulator(net, seed=seed) as sim0:
        sim0.step()

    if not eager_enabled():
        # this is necessary to reset the graph seed
        tf.keras.backend.clear_session()

    with Simulator(net, seed=seed) as sim1:
        sim1.step()

    assert np.allclose(sim0.data[p], sim1.data[p])


@pytest.mark.xfail(reason="TensorFlow does not support resetting RNG")
def test_tf_rng_reset(Simulator, seed):
    # TODO: support this using tf.random.experimental; see
    #  https://github.com/nengo/nengo-dl/issues/98

    with nengo.Network() as net:
        a = TensorNode(lambda t: tf.random.uniform((1, 1), dtype=t.dtype))
        p = nengo.Probe(a)

    with Simulator(net, seed=seed) as sim:
        sim.step()

        data = sim.data[p]

        sim.reset(seed=seed + 1)
        sim.step()
        assert not np.allclose(sim.data[p], data)

        sim.reset(seed=seed)
        sim.step()
        assert np.allclose(sim.data[p], data)


def test_pickle_error(Simulator):
    with pytest.raises(NotImplementedError, match="does not support pickling"):
        pickle.dumps(Simulator(nengo.Network()))


def test_tensorflow_gpu_warning(Simulator, pytestconfig):
    with pytest.warns(None) as recwarns:
        with Simulator(nengo.Network()):
            pass

    recwarns = [w for w in recwarns if "No GPU support detected" in str(w.message)]

    assert len(recwarns) == (
        1
        if not utils.tf_gpu_installed and pytestconfig.getoption("--device") is None
        else 0
    )


def test_train_loss_deprecation(Simulator):
    with Simulator(dummies.linear_net()[0]) as sim:
        with pytest.raises(SimulationError, match="train has been deprecated"):
            sim.train()

        with pytest.raises(SimulationError, match="loss has been deprecated"):
            sim.loss()


def test_io_names(Simulator):
    with nengo.Network() as net:
        nodes = [
            nengo.Node([0]),
            nengo.Node([0]),
            nengo.Node([0], label="aa"),
            nengo.Node([0], label="aa"),
            nengo.Node([0], label="Aa"),
            nengo.Node([0], label="aA"),
            nengo.Node([0]),
        ]

        for n in nodes:
            nengo.Probe(n, label=None if n.label is None else "p_" + n.label)

    with Simulator(net) as sim:
        assert sim.keras_model.input_names == [
            "node",
            "node_1",
            "aa",
            "aa_1",
            "Aa_2",
            "aA_3",
            "node_2",
            "n_steps",
        ]
        assert sim.keras_model.output_names == [
            "probe",
            "probe_1",
            "p_aa",
            "p_aa_1",
            "p_Aa_2",
            "p_aA_3",
            "probe_2",
            "steps_run",
        ]
        sim.step()

        with pytest.raises(ValidationError, match="not an input Node"):
            sim.get_name(nengo.Node([0], add_to_container=False))

        with pytest.raises(ValidationError, match="from a different network"):
            sim.get_name(nengo.Probe(nodes[0], add_to_container=False))

        with pytest.raises(ValidationError, match="unknown type"):
            sim.get_name(nengo.Ensemble(10, 1, add_to_container=False))


def test_log_filter(Simulator, caplog):
    with nengo.Network() as net:
        a = nengo.Node([0])
        p = nengo.Probe(a)

    with Simulator(net) as sim:
        sim.compile(loss={p: "mse"})
        assert "missing from loss dictionary" not in caplog.text

        sim.tensor_graph.add_weight()
        assert "constraint is deprecated" not in caplog.text

        tf.compat.v1.keras.backend.get_session()
        assert "tf.keras.backend.get_session" not in caplog.text

        # make sure there were no other deprecation warnings
        filt = TFLogFilter(err_on_deprecation=True)
        for rec in caplog.records:
            filt.filter(rec)

    # make sure the deprecation checking works
    logger = logging.getLogger("test")
    logger.addFilter(filt)
    with pytest.raises(Exception, match="Deprecation warning"):
        logger.warning("a thing is deprecated")


@pytest.mark.training
def test_uneven_batch_size(Simulator):
    with nengo.Network() as net:
        inp = nengo.Node(np.zeros(2))
        nengo.Probe(inp)

    with Simulator(net, minibatch_size=16) as sim:
        with pytest.warns(UserWarning, match="not evenly divisible"):
            sim.predict(np.zeros((20, 5, 2)))

        sim.compile(optimizer=tf.optimizers.SGD(0), loss=tf.losses.mse)

        with pytest.warns(UserWarning, match="not evenly divisible"):
            sim.fit(np.zeros((20, 5, 2)), np.zeros((20, 5, 2)))


def test_progress_bar(Simulator, capsys, monkeypatch):
    class StdProgressBar(utils.ProgressBar):  # pylint: disable=too-many-ancestors
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            # manually set fd to stdout (passing `fd=sys.stdout` in init is not the
            # same, because it wraps it in some stream wrapper that is slightly
            # different). we want to use sys.stdout exactly so that stdout capturing
            # (e.g. in capsys) works correctly.
            self.fd = sys.stdout

    monkeypatch.setattr(utils, "ProgressBar", StdProgressBar)

    net, _, _ = dummies.linear_net()

    with Simulator(net, progress_bar=True) as sim:
        # default to displaying build information
        build_output = capsys.readouterr().out
        assert "Building" in build_output
        assert "Optimization" in build_output
        assert "Construction" in build_output

        # display simulation progress
        sim.run_steps(10)
        assert "Simulating" in capsys.readouterr().out

        # simulation progress explicitly disabled
        sim.run_steps(10, progress_bar=False)
        assert capsys.readouterr().out == ""

    with Simulator(net, progress_bar=False) as sim:
        # no build information
        assert capsys.readouterr().out == ""

        # defaults to false
        sim.run_steps(10)
        assert capsys.readouterr().out == ""

        # can be explicitly enabled
        sim.run_steps(10, progress_bar=True)
        assert "Simulating" in capsys.readouterr().out


def test_soft_reset(Simulator):
    with Simulator(nengo.Network()) as sim:
        sim.run_steps(10)
        assert sim.n_steps == 10
        with pytest.warns(DeprecationWarning, match="use Simulator.reset"):
            sim.soft_reset()
        assert sim.n_steps == 0


def test_sim_close(Simulator):
    with Simulator(nengo.Network()) as sim:
        assert sim.tensor_graph

    with pytest.raises(SimulatorClosed, match="access Simulator.tensor_graph"):
        assert sim.tensor_graph

    with pytest.raises(SimulatorClosed, match="simulator is closed"):
        with sim:
            pass


def test_logging(Simulator, caplog):
    with nengo.Network() as net:
        inp = nengo.Node([0])
        ens = nengo.Ensemble(10, 1)
        nengo.Connection(inp, ens)
        nengo.Probe(ens)

    # run a simulation with logging to verify that there are no errors
    with caplog.at_level(logging.NOTSET):
        with Simulator(net) as sim:
            sim.run_steps(10)

    for rec in caplog.records:
        assert rec.getMessage(), "Record %s has empty message" % rec


def test_floatx_context(Simulator):
    with nengo.Network() as net:
        configure_settings(dtype="float64")

        def fail_func(t):
            assert tf.keras.backend.floatx() == "float64"
            assert False, "intentional failure"

        nengo.Node(fail_func, size_out=1)

    assert tf.keras.backend.floatx() == "float32"
    sim = Simulator(net)
    with pytest.raises(AssertionError, match="intentional failure"):
        sim.step()
    assert tf.keras.backend.floatx() == "float32"
    sim.close()



"""Tools for automatically converting a Keras model to a Nengo network."""

import collections
import copy
import logging
import warnings

import nengo
import numpy as np
import tensorflow as tf
from packaging import version
from tensorflow.python.keras import engine
from tensorflow.python.keras.layers import BatchNormalizationV1, BatchNormalizationV2

from nengo_dl import compat
from nengo_dl.config import configure_settings
from nengo_dl.neurons import LeakyReLU
from nengo_dl.simulator import Simulator
from nengo_dl.tensor_node import Layer, TensorNode

logger = logging.getLogger(__name__)


class Converter:
    """
    Converts a Keras model to a Nengo network composed of native Nengo objects.

    Parameters
    ----------
    model : ``tf.keras.Model``
        Keras model to be converted
    allow_fallback : bool
        If True, allow layers that cannot be converted to native Nengo
        objects to be added as a `.TensorNode` instead. Note that if this occurs, the
        converted Nengo network will only be runnable in the NengoDL simulator.
    inference_only : bool
        Allow layers to be converted in such a way that inference behaviour will
        match the source model but training behaviour will not.
        If ``inference_only=False`` then some
        layers cannot be converted to native Nengo objects (but you can
        still use ``allow_fallback=True`` to use a `.TensorNode` instead).
    max_to_avg_pool : bool
        If True, convert max pooling layers to average pooling layers. Note that this
        will change the behaviour of the network, so parameters will probably need to
        be re-trained in NengoDL. If ``max_to_avg_pool=False`` then max pooling layers
        cannot be converted to native Nengo objects (but you can
        still use ``allow_fallback=True`` to use a `.TensorNode` instead).
    split_shared_weights : bool
        In Keras, applying the same ``Layer`` object to different input layers will
        result in multiple instances of the given layer that share the same weights.
        This is not supported in Nengo. If ``split_shared_weights=True`` then those
        shared weights will be split into independent sets of weights. They will all
        be initialized to the same value, so the initial behaviour of the model will
        be unchanged, but if any further training is performed on the network then
        the weights in each of those instances may diverge.
    swap_activations : dict
        A dictionary mapping from TensorFlow activation functions or Nengo neuron types
        to TensorFlow activation functions or Nengo neuron types. This can be used to
        change all the activation types in a model to some other type. This is in
        addition to the default activation map (see `.LayerConverter`). It can be keyed
        based on either TensorFlow or Nengo activation types, and will be applied both
        before and after the default activation map, in order to support whatever swap
        type is most useful for a given model. In particular, ``swap_activations``
        can be useful for swapping rate neuron types to spiking neuron types,
        through e.g. ``{tf.nn.relu: nengo.SpikingRectifiedLinear()}`` or
        ``{nengo.RectifiedLinear(): nengo.SpikingRectifiedLinear()}``.  Or it can be
        used to swap activation types that don't have a native Nengo implementation,
        e.g. ``{tf.keras.activatons.elu: tf.keras.activations.relu}``.
    scale_firing_rates: float or dict
        Scales the inputs of neurons by ``x``, and the outputs by ``1/x``.
        The idea is that this parameter can be used to increase the firing rates of
        spiking neurons (by scaling the input), without affecting the overall output
        (because the output spikes are being scaled back down). Note that this is only
        strictly true for neuron types with linear activation functions (e.g. ReLU).
        Nonlinear neuron types (e.g. LIF) will be skewed by this linear scaling on the
        input/output. ``scale_firing_rates`` can be specified as a float, which will
        be applied to all layers in the model, or as a dictionary mapping Keras model
        layers to a scale factor, allowing different scale factors to be applied to
        different layers.
    synapse : float or `nengo.synapses.Synapse`
        Synaptic filter to be applied on the output of all neurons. This can be useful
        to smooth out the noise introduced by spiking neurons. Note, however, that
        increasing the synaptic filtering will make the network less responsive to
        rapid changes in the input, and you may need to present each input value
        for more timesteps in order to allow the network output to settle.
    temporal_model : bool
        Set to True if the Keras model contains a temporal
        dimension (e.g. the inputs and outputs of each layer have shape
        ``(batch_size, n_steps, ...)``). Note that all layers must be temporal, the
        Converter cannot handle models with a mix of temporal and non-temporal layers.

    Attributes
    ----------
    model : ``tf.keras.Model``
        The input Keras model (if input was a Sequential model then this will be the
        equivalent Functional model).
    net : `nengo.Network`
        The converted Nengo network.
    inputs : `.Converter.KerasTensorDict`
        Maps from Keras model inputs to input Nodes in the converted Nengo network.
        For example, ``my_node = Converter(my_model).inputs[my_model.input]``.
    outputs : `.Converter.KerasTensorDict`
        Maps from Keras model outputs to output Probes in the converted Nengo network.
        For example, ``my_probe = Converter(my_model).outputs[my_model.output]``.
    layers : `.Converter.KerasTensorDict`
        Maps from Keras model layers to the converted Nengo object.
        For example, ``my_neurons = Converter(my_model).layers[my_layer]``.
    """

    converters = {}

    def __init__(
        self,
        model,
        allow_fallback=True,
        inference_only=False,
        max_to_avg_pool=False,
        split_shared_weights=False,
        swap_activations=None,
        scale_firing_rates=None,
        synapse=None,
        temporal_model=False,
    ):
        self.allow_fallback = allow_fallback
        self.inference_only = inference_only
        self.max_to_avg_pool = max_to_avg_pool
        self.split_shared_weights = split_shared_weights
        self.swap_activations = Converter.TrackedDict(swap_activations or {})
        self.scale_firing_rates = scale_firing_rates
        self.synapse = synapse
        self.temporal_model = temporal_model
        self._layer_converters = {}

        # convert model
        self.net = self.get_converter(model).convert(None)

        # set model from the converter in case the converter has changed the model type
        # (i.e. if the model is sequential, it will be converted to a functional model)
        self.model = self.get_converter(model).layer

        if self.swap_activations.unused_keys():
            warnings.warn(
                "swap_activations contained %s, but there were no layers in the model "
                "with that activation type" % (self.swap_activations.unused_keys(),)
            )

        self.layers = self.net.layers

        with self.net:
            configure_settings(inference_only=self.inference_only)

            # convert inputs to input nodes
            self.inputs = Converter.KerasTensorDict()
            for input in self.model.inputs:
                input_obj = self.net.inputs[input]
                logger.info("Setting input node %s (%s)", input_obj, input)
                input_size = input_obj.size_in
                input_obj.size_in = 0
                input_obj.output = np.zeros(input_size)
                self.inputs[input] = input_obj

            # add probes to outputs
            self.outputs = Converter.KerasTensorDict()
            for output in self.model.outputs:
                output_obj = self.net.outputs[output]
                logger.info("Probing %s (%s)", output_obj, output)
                self.outputs[output] = nengo.Probe(
                    output_obj,
                    synapse=self.synapse
                    if isinstance(output_obj, nengo.ensemble.Neurons)
                    else None,
                )

    def verify(self, training=False, inputs=None, atol=1e-8, rtol=1e-5):
        """
        Verify that output of converted Nengo network matches the original Keras model.

        Parameters
        ----------
        training : bool
            If True, check that optimizing the converted Nengo network produces the same
            results as optimizing the original Keras model.
        inputs : list of `numpy.ndarray`
            Testing values for model inputs (if not specified, array of ones will be
            used).
        atol : float
            Absolute tolerance for difference between Nengo and Keras outputs.
        rtol : float
            Relative (to Nengo) tolerance for difference between nengo and Keras
            outputs.

        Returns
        -------
        success : bool
            True if output of Nengo network matches output of Keras model.

        Raises
        ------
        ValueError
            If output of Nengo network does not match output of Keras model.
        """

        epochs = 3
        batch_size = 2 if inputs is None else inputs[0].shape[0]
        n_steps = (
            1
            if not self.temporal_model
            else (10 if inputs is None else inputs[0].shape[1])
        )
        shape_prefix = (batch_size, n_steps) if self.temporal_model else (batch_size,)

        inp_vals = (
            [
                np.ones(shape_prefix + x.shape[len(shape_prefix) :])
                for x in self.model.inputs
            ]
            if inputs is None
            else inputs
        )

        # get keras model output
        if training:
            out_vals = [
                np.ones(shape_prefix + x.shape[len(shape_prefix) :])
                for x in self.model.outputs
            ]
            self.model.compile(optimizer=tf.optimizers.SGD(0.1), loss=tf.losses.mse)
            self.model.fit(inp_vals, out_vals, epochs=epochs)

        keras_out = self.model.predict(inp_vals)

        if not isinstance(keras_out, (list, tuple)):
            keras_out = [keras_out]

        # get nengo sim output
        inp_vals = [np.reshape(x, (batch_size, n_steps, -1)) for x in inp_vals]

        with Simulator(self.net, minibatch_size=batch_size) as sim:
            if training:
                keras_params = sum(
                    np.prod(w.shape) for w in self.model.trainable_weights
                )
                nengo_params = sum(
                    np.prod(w.shape) for w in sim.keras_model.trainable_weights
                )
                if keras_params != nengo_params:
                    raise ValueError(
                        "Number of trainable parameters in Nengo network (%d) does not "
                        "match number of trainable parameters in Keras model (%d)"
                        % (nengo_params, keras_params)
                    )

                out_vals = [np.reshape(x, (batch_size, n_steps, -1)) for x in out_vals]
                sim.compile(optimizer=tf.optimizers.SGD(0.1), loss=tf.losses.mse)
                sim.fit(inp_vals, out_vals, epochs=epochs)

            sim_out = sim.predict(inp_vals)

        for i, out in enumerate(self.model.outputs):
            keras_vals = np.ravel(keras_out[i])
            nengo_vals = np.ravel(sim_out[self.outputs[out]])
            fails = np.logical_not(
                np.isclose(keras_vals, nengo_vals, atol=atol, rtol=rtol)
            )
            if np.any(fails):
                logger.info("Verification failure")
                logger.info("Keras:\n%s", keras_vals[fails])
                logger.info("Nengo:\n%s", nengo_vals[fails])
                raise ValueError(
                    "Output of Keras model does not match output of converted "
                    "Nengo network (max difference=%.2E; set log level to INFO to see "
                    "all failures)" % max(abs(keras_vals[fails] - nengo_vals[fails]))
                )

        return True

    def get_converter(self, layer):
        """
        Get instantiated `.LayerConverter` for the given ``Layer`` instance.

        Note that this caches the results, so calling the function multiple times
        with the same Layer instance will return the same LayerConverter instance.

        Parameters
        ----------
        layer : ``tf.keras.layers.Layer``
            The Keras Layer being converted.

        Returns
        -------
        converter : `.LayerConverter`
            LayerConverter class for converting ``layer`` to Nengo objects.
        """

        if layer in self._layer_converters:
            # already have an instantiated converter for this layer
            converter = self._layer_converters[layer]
            if converter.has_weights and not self.split_shared_weights:
                # TODO: allow fallback
                raise ValueError(
                    "Multiple applications of layer %s detected; this is not supported "
                    "unless split_shared_weights=True" % layer
                )
            return converter

        # check if there is a registered builder for this layer type
        ConverterClass = self.converters.get(type(layer), None)

        # perform custom checks in layer converters
        if ConverterClass is None:
            convertible = False
            error_msg = "Layer type %s does not have a registered converter" % type(
                layer
            )
        else:
            convertible, error_msg = ConverterClass.convertible(layer, self)

        if not convertible:
            # this means that there is no LayerConverter compatible with this layer
            # (either because it has an unknown type, or it failed the ``.convertible``
            # check due to its internal parameterization)
            can_fallback = ConverterClass is None or ConverterClass.allow_fallback
            if self.allow_fallback and can_fallback:
                warnings.warn(
                    "%sFalling back to TensorNode."
                    % (error_msg + ". " if error_msg else "")
                )
                ConverterClass = self.converters[None]
            else:
                msg = "%sUnable to convert layer %s to native Nengo objects; " % (
                    error_msg + ". " if error_msg else "",
                    layer.name,
                )
                if not self.allow_fallback and can_fallback:
                    msg += (
                        "set allow_fallback=True if you would like to use a "
                        "TensorNode instead, or "
                    )
                msg += (
                    "consider registering a custom LayerConverter for this layer type."
                )
                raise TypeError(msg)

        converter = ConverterClass(layer, self)

        self._layer_converters[layer] = converter

        return converter

    @classmethod
    def register(cls, keras_layer):
        """
        A decorator for adding a class to the converter registry.

        Parameters
        ----------
        keras_layer : ``tf.keras.layers.Layer``
            The Layer associated with the conversion function being registered.
        """

        def register_converter(convert_cls):
            if keras_layer in cls.converters:
                warnings.warn(
                    "Layer '%s' already has a converter. Overwriting." % keras_layer
                )

            cls.converters[keras_layer] = convert_cls

            return convert_cls

        return register_converter

    class KerasTensorDict(collections.abc.Mapping):
        """
        A dictionary-like object that has extra logic to handle Layer/Tensor keys.
        """

        def __init__(self):
            self.dict = collections.OrderedDict()

        def _get_key(self, key):
            if isinstance(key, tf.keras.layers.Layer):
                if len(key.inbound_nodes) > 1:
                    raise KeyError(
                        "Layer %s is ambiguous because it has been called multiple "
                        "times; use a specific set of layer outputs as key instead"
                        % key
                    )

                # get output tensor
                key = key.output

            key = tuple(x.ref() if tf.is_tensor(x) else x for x in tf.nest.flatten(key))

            return key

        def __setitem__(self, key, val):
            self.dict[self._get_key(key)] = val

        def __getitem__(self, key):
            return self.dict[self._get_key(key)]

        def __iter__(self):
            return iter(self.dict)

        def __len__(self):
            return len(self.dict)

    class TrackedDict(collections.abc.Mapping):
        """
        A dictionary-like object that keeps track of which keys have been accessed.
        """

        def __init__(self, dict):
            self.dict = dict
            self.read = set()

        def __getitem__(self, key):
            self.read.add(key)
            return self.dict[key]

        def __iter__(self):
            return iter(self.dict)

        def __len__(self):
            return len(self.dict)

        def unused_keys(self):
            """Returns any keys in the dictionary that have never been read."""
            return set(self.dict.keys()) - self.read


class LayerConverter:
    """
    Base class for converter classes, which contain the logic for mapping some Keras
    layer type to Nengo objects.

    Subclasses must implement the `.LayerConverter.convert` method. They may optionally
    extend `.LayerConverter.convertible` if this layer type requires custom logic for
    whether or not a layer can be converted to Nengo objects.

    Subclasses should override the ``unsupported_args`` class parameter if there are
    certain non-default Layer attributes that are not supported by the converter.
    This is a list of names for attributes that must have the default value for the
    layer to be convertible. The default is assumed to be ``None``, or a tuple of
    ``("attribute_name", default_value)`` can be specified. If there are parameters
    that are supported in inference mode but not in training mode, they should be
    added to the ``unsupported_training_args`` parameter.

    Subclasses should override the ``has_weights`` class parameter if the layer type
    being converted contains internal weights (this affects how the converter will
    handle duplicate layers).

    Parameters
    ----------
    layer : ``tf.keras.layers.Layer``
        The Layer object being converted.
    converter : `.Converter`
        The parent Converter class running the conversion process.
    """

    # maps from TensorFlow activation functions to Nengo neuron types
    activation_map = {
        None: None,
        tf.keras.activations.linear: None,
        tf.keras.activations.relu: nengo.RectifiedLinear(),
        tf.nn.relu: nengo.RectifiedLinear(),
        tf.keras.activations.sigmoid: nengo.Sigmoid(tau_ref=1),
        tf.nn.sigmoid: nengo.Sigmoid(tau_ref=1),
    }
    if version.parse(nengo.__version__) > version.parse("3.0.0"):
        activation_map.update(
            {
                tf.keras.activations.tanh: compat.Tanh(tau_ref=1),
                tf.nn.tanh: compat.Tanh(tau_ref=1),
            }
        )

    # attributes of the Keras layer that are not supported for non-default values.
    # the default value is assumed to be None, or a tuple of
    # ("attr_name", default_value) can be specified
    unsupported_args = []
    # attributes that are supported in inference_only mode but otherwise not
    unsupported_training_args = []

    # whether or not this layer contains trainable weights (this indicates whether
    # this layer is affected by split_shared_weights)
    has_weights = False

    # whether or not this layer supports a TensorNode fallback
    allow_fallback = True

    def __init__(self, layer, converter):
        self.layer = layer
        self.converter = converter

    def add_nengo_obj(self, node_id, biases=None, activation=None):
        """
        Builds a Nengo object for the given Node of this layer.

        Parameters
        ----------
        node_id : int
            The index of the Keras Node currently being built on this layer.
        biases : `numpy.ndarray` or None
            If not None, add trainable biases with the given value.
        activation : callable or None
            The TensorFlow activation function to be used (``None`` will be
            interpreted as linear activation).

        Returns
        -------
        obj : `nengo.Node` or `nengo.ensemble.Neurons` or `nengo_dl.TensorNode`
            The Nengo object whose output corresponds to the output of the given Keras
            Node.
        """
        name = self.layer.name + ".%d" % node_id

        # apply manually specified swaps
        activation = self.converter.swap_activations.get(activation, activation)

        if activation in self.activation_map or isinstance(
            activation, nengo.neurons.NeuronType
        ):
            activation = self.activation_map.get(activation, activation)

            # apply any nengo->nengo swaps
            activation = self.converter.swap_activations.get(activation, activation)

            if activation is None:
                # linear activation, uses a passthrough Node
                obj = nengo.Node(
                    size_in=np.prod(self.output_shape(node_id)), label=name
                )
            else:
                # use ensemble to implement the appropriate neuron type

                # apply firing rate scaling
                if self.converter.scale_firing_rates is None:
                    scale_firing_rates = None
                elif isinstance(self.converter.scale_firing_rates, dict):
                    # look up specific layer rate
                    scale_firing_rates = self.converter.scale_firing_rates.get(
                        self.layer, None
                    )
                else:
                    # constant scale applied to all layers
                    scale_firing_rates = self.converter.scale_firing_rates

                if scale_firing_rates is None:
                    gain = 1
                else:
                    gain = scale_firing_rates
                    if biases is not None:
                        biases *= scale_firing_rates

                    if hasattr(activation, "amplitude"):
                        # copy activation so that we can change amplitude without
                        # affecting other instances
                        activation = copy.copy(activation)

                        # bypass read-only protection
                        type(activation).amplitude.data[
                            activation
                        ] /= scale_firing_rates
                    else:
                        warnings.warn(
                            "Firing rate scaling being applied to activation type "
                            "that does not support amplitude (%s); this will change "
                            "the output" % type(activation)
                        )

                obj = nengo.Ensemble(
                    np.prod(self.output_shape(node_id)),
                    1,
                    neuron_type=activation,
                    gain=nengo.dists.Choice([gain]),
                    bias=nengo.dists.Choice([0]) if biases is None else biases,
                    label=name,
                ).neurons
                if biases is None:
                    # ensembles always have biases, so if biases=None we just use
                    # all-zero biases and mark them as non-trainable
                    self.set_trainable(obj, False)
        elif self.converter.allow_fallback:
            warnings.warn(
                "Activation type %s does not have a native Nengo equivalent; "
                "falling back to a TensorNode" % activation
            )
            obj = TensorNode(
                activation,
                shape_in=self.output_shape(node_id),
                pass_time=False,
                label=name,
            )
        else:
            raise TypeError("Unsupported activation type (%s)" % self.layer.activation)

        if biases is not None and isinstance(obj, (nengo.Node, TensorNode)):
            # obj doesn't have its own biases, so use a connection from a constant node
            # (so that the bias values will be trainable)
            bias_node = nengo.Node([1], label="%s.bias" % name)
            nengo.Connection(bias_node, obj, transform=biases[:, None], synapse=None)

        logger.info("Created %s (size=%d)", obj, obj.size_out)

        return obj

    def add_connection(self, node_id, obj, input_idx=0, trainable=False, **kwargs):
        """
        Adds a Connection from one of the inputs of the Node being built to the
        Nengo object.

        Parameters
        ----------
        node_id : int
            The index of the Keras Node currently being built on this layer.
        obj : ``NengoObject``
            The Nengo object implementing this Node.
        input_idx : int
            Which of the inputs we want to add a Connection for (in the case of
            layers that have multiple inputs).
        trainable : bool
            Whether or not the weights associated with the created Connection
            should be trainable.
        kwargs : dict
            Will be passed on to `nengo.Connection`.

        Returns
        -------
        conn : `nengo.Connection`
            The constructed Connection object.
        """

        pre = self.get_input_obj(node_id, tensor_idx=input_idx)

        kwargs.setdefault(
            "synapse",
            self.converter.synapse if isinstance(pre, nengo.ensemble.Neurons) else None,
        )

        conn = nengo.Connection(
            pre,
            obj,
            **kwargs,
        )
        self.set_trainable(conn, trainable)

        logger.info("Connected %s to %s (trainable=%s)", conn.pre, conn.post, trainable)

        return conn

    def get_input_obj(self, node_id, tensor_idx=0):
        """
        Returns the Nengo object corresponding to the given input of this layer.

        Parameters
        ----------
        node_id : int
            The index of the Keras Node currently being built on this layer.
        tensor_idx : int
            The index of the input we want to look up (for layers with multiple inputs).

        Returns
        -------
        obj : `nengo.Node` or `nengo.ensemble.Neurons` or `nengo_dl.TensorNode`
            The Nengo object whose output corresponds to the given input of this layer.
        """
        input_node = self.layer.inbound_nodes[node_id]

        input_tensors = input_node.input_tensors

        if isinstance(input_tensors, (list, tuple)):
            tensor = input_tensors[tensor_idx]
        else:
            assert tensor_idx == 0
            tensor = input_tensors

        input_layer, input_node_id, input_tensor_id = tensor._keras_history

        return nengo.Network.context[-1].layer_map[input_layer][input_node_id][
            input_tensor_id
        ]

    def _get_shape(self, input_output, node_id, full_shape=False):
        """
        Looks up the input or output shape of this Node.

        Parameters
        ----------
        input_output : "input" or "output"
            Whether we want the input or output shape.
        node_id : int
            The node whose shape we want to look up.
        full_shape : bool
            Whether or not the returned shape should include the batch/time dimensions.

        Returns
        -------
        shape : (list of) tuple of int
            A single tuple shape if the node has one input/output, or a list of shapes
            if the node as multiple inputs/outputs.
        """

        # note: layer.get_input/output_shape_at is generally equivalent to
        # layer.input/output_shape, except when the layer is called multiple times
        # with different shapes, in which case input/output_shape is not well defined
        func = getattr(self.layer, "get_%s_shape_at" % input_output)

        # get the shape
        shape = func(node_id)

        if not full_shape:
            to_skip = 2 if self.converter.temporal_model else 1
            if isinstance(shape, list):
                # multiple inputs/outputs; trim the batch from each one
                shape = [s[to_skip:] for s in shape]
            else:
                shape = shape[to_skip:]

        return shape

    def input_shape(self, node_id, full_shape=False):
        """
        Returns the input shape of the given node.

        Parameters
        ----------
        node_id : int
            The node whose shape we want to look up.
        full_shape : bool
            Whether or not the returned shape should include the batch/time dimensions.

        Returns
        -------
        shape : (list of) tuple of int
            A single tuple shape if the node has one input, or a list of shapes
            if the node as multiple inputs.
        """
        return self._get_shape("input", node_id, full_shape=full_shape)

    def output_shape(self, node_id, full_shape=False):
        """
        Returns the output shape of the given node.

        Parameters
        ----------
        node_id : int
            The node whose shape we want to look up.
        full_shape : bool
            Whether or not the returned shape should include the batch/time dimensions.

        Returns
        -------
        shape : (list of) tuple of int
            A single tuple shape if the node has one output, or a list of shapes
            if the node as multiple outputs.
        """
        return self._get_shape("output", node_id, full_shape=full_shape)

    @staticmethod
    def set_trainable(obj, trainable):
        """
        Set trainable config attribute of object.

        Parameters
        ----------
        obj : ``NengoObject``
            The object to be assigned.
        trainable: bool
            Trainability value of ``obj``.
        """

        nengo.Network.context[-1].config[obj].trainable = trainable

    @classmethod
    def convertible(cls, layer, converter):
        """
        Check whether the given Keras layer is convertible to native Nengo objects.

        Parameters
        ----------
        layer : ``tf.keras.layers.Layer``
            The Keras Layer we want to convert.
        converter : `.Converter`
            The Converter object running the conversion process.

        Returns
        -------
        convertible : bool
            True if the layer can be converted to native Nengo objects, else False.
        """
        # check if the layer uses any unsupported arguments
        unsupported = cls.unsupported_args
        if not converter.inference_only:
            unsupported = unsupported + cls.unsupported_training_args
        for arg in unsupported:
            if isinstance(arg, str):
                default = None
            else:
                arg, default = arg

            val = getattr(layer, arg)
            if val != default:
                msg = "%s.%s has value %s != %s, which is not supported" % (
                    layer.name,
                    arg,
                    val,
                    default,
                )
                if arg in cls.unsupported_training_args:
                    msg += " (unless inference_only=True)"
                return False, msg

        return True, None

    def convert(self, node_id):
        """
        Convert the given node of this layer to Nengo objects

        Parameters
        ----------
        node_id : int
            The index of the inbound node to be converted.

        Returns
        -------
        output : ``NengoObject``
            Nengo object whose output corresponds to the output of the Keras layer node.
        """
        raise NotImplementedError("Subclasses must implement convert")


@Converter.register(
    tf.keras.Model
    if version.parse(tf.__version__) < version.parse("2.3.0rc0")
    else engine.functional.Functional
)
class ConvertModel(LayerConverter):
    """Convert ``tf.keras.Model`` to Nengo objects."""

    def convert(self, node_id):
        logger.info("=" * 30)
        logger.info("Converting model %s", self.layer.name)

        # functional models should already have been built when the model
        # was instantiated
        assert self.layer.built

        # trace the model to find all the nodes that need to be built into
        # the Nengo network
        ordered_nodes, _ = compat._build_map(self.layer.outputs)
        layer_ids = [
            (node.layer, node.layer.inbound_nodes.index(node)) for node in ordered_nodes
        ]

        with nengo.Network(
            label=self.layer.name + ("" if node_id is None else ".%d" % node_id)
        ) as net:
            # add the "trainable" attribute to all objects
            configure_settings(trainable=None)

            net.layer_map = collections.defaultdict(dict)

            for layer, layer_node_id in layer_ids:
                # should never be rebuilding the same node
                assert layer_node_id not in net.layer_map[layer]

                logger.info("-" * 30)
                logger.info("Converting layer %s node %d", layer.name, layer_node_id)

                # get the layerconverter object
                layer_converter = self.converter.get_converter(layer)

                # build the Nengo objects
                nengo_layer = layer_converter.convert(layer_node_id)
                assert isinstance(
                    nengo_layer,
                    (nengo.Node, nengo.ensemble.Neurons, TensorNode, nengo.Network),
                )

                # add output of layer_converter to layer_map
                net.layer_map[layer][layer_node_id] = (
                    list(nengo_layer.outputs.values())
                    if isinstance(nengo_layer, nengo.Network)
                    else [nengo_layer]
                )

            # data structures to track converted objects
            net.inputs = Converter.KerasTensorDict()
            for input in self.layer.inputs:
                input_layer, input_node_id, input_tensor_id = input._keras_history
                net.inputs[input] = net.layer_map[input_layer][input_node_id][
                    input_tensor_id
                ]

            net.outputs = Converter.KerasTensorDict()
            for output in self.layer.outputs:
                output_layer, output_node_id, output_tensor_id = output._keras_history
                net.outputs[output] = net.layer_map[output_layer][output_node_id][
                    output_tensor_id
                ]

            net.layers = Converter.KerasTensorDict()
            for layer in self.layer.layers:
                for layer_node_id, node_outputs in net.layer_map[layer].items():
                    for nengo_obj in node_outputs:
                        output_tensors = layer.inbound_nodes[
                            layer_node_id
                        ].output_tensors
                        net.layers[output_tensors] = nengo_obj

        if node_id is not None:
            # add incoming connections in parent network
            for i, inp in enumerate(net.inputs.values()):
                self.add_connection(node_id, inp, input_idx=i)

        logger.info("=" * 30)

        return net


@Converter.register(tf.keras.Sequential)
class ConvertSequential(ConvertModel):
    """Convert ``tf.keras.Sequential`` to Nengo objects."""

    def __init__(self, seq_model, converter):
        # convert sequential model to functional model
        warnings.warn(
            "Converting sequential model to functional model; "
            "use `Converter.model` to refer to the functional model (rather than the "
            "original sequential model) when working with the output of the Converter"
        )
        input_shape = seq_model.layers[0].input_shape

        inp = x = tf.keras.Input(batch_shape=input_shape)
        for layer in seq_model.layers:
            x = layer(x)

        func_model = tf.keras.Model(inp, x)

        super().__init__(func_model, converter)


@Converter.register(None)
class ConvertFallback(LayerConverter):
    """
    Convert layers which do not have a native Nengo equivalent into a
    `.TensorNode`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # copy layer, so that any changes to the layer (e.g., rebuilding it), do not
        # affect the source model
        layer = self.layer.__class__.from_config(self.layer.get_config())
        if self.layer.built:
            layer.build(self.input_shape(0, full_shape=True))
            layer.set_weights(self.layer.get_weights())

        self.tensor_layer = Layer(layer)

    def convert(self, node_id):
        logger.info("Using TensorNode %s", self.tensor_layer)

        input_obj = self.get_input_obj(node_id)
        output = self.tensor_layer(
            input_obj, shape_in=self.input_shape(node_id), label=self.layer.name
        )

        logger.info("Applying to %s, created %s", input_obj, output)

        return output


class ConvertAvgPool(LayerConverter):
    """Base class for converting average pooling layers to Nengo objects."""

    # "same" padding not supported because TensorFlow average pooling does not count
    # the padded cells in the average, which we don't have a way to do using Convolution
    unsupported_args = [("padding", "valid")]

    def convert(self, node_id, dimensions):
        output = self.add_nengo_obj(node_id)

        def to_tuple(val):
            if isinstance(val, int):
                return (val,) * dimensions

            return val

        spatial_shape = (
            self.input_shape(node_id)[:-1]
            if self.layer.data_format == "channels_last"
            else self.input_shape(node_id)[1:]
        )

        # the default values here are for GlobalAveragePooling (which doesn't have
        # these attributes)
        pool_size = to_tuple(getattr(self.layer, "pool_size", spatial_shape))
        padding = getattr(self.layer, "padding", "valid")
        strides = to_tuple(getattr(self.layer, "strides", 1))

        # the idea here is that we set up a convolutional transform with weights 1/n,
        # which will have the effect of implementing average pooling
        n_filters = (
            self.output_shape(node_id)[-1]
            if self.layer.data_format == "channels_last"
            else self.output_shape(node_id)[0]
        )
        n_pool = np.prod(pool_size)
        kernel = np.reshape(
            [np.eye(n_filters) / n_pool] * n_pool, pool_size + (n_filters, n_filters)
        )
        pool_conv = nengo.Convolution(
            n_filters=n_filters,
            input_shape=self.input_shape(node_id),
            padding=padding,
            strides=strides,
            kernel_size=pool_size,
            init=kernel,
            channels_last=self.layer.data_format == "channels_last",
        )
        self.add_connection(node_id, output, transform=pool_conv)

        return output

    @classmethod
    def convertible(cls, layer, converter):
        if (
            isinstance(
                layer,
                (
                    tf.keras.layers.MaxPool1D,
                    tf.keras.layers.MaxPool2D,
                    tf.keras.layers.MaxPool3D,
                    tf.keras.layers.GlobalMaxPool1D,
                    tf.keras.layers.GlobalMaxPool2D,
                    tf.keras.layers.GlobalMaxPool3D,
                ),
            )
            and not converter.max_to_avg_pool
        ):
            msg = (
                "Cannot convert max pooling layers to native Nengo objects; consider "
                "setting max_to_avg_pool=True to use average pooling instead"
            )
            return False, msg

        unsupported = cls.unsupported_args
        if not hasattr(layer, "padding"):
            # global layers don't have this attribute, so we temporarily remove it
            # from the unsupported args
            cls.unsupported_args = []

        try:
            convertible = super().convertible(layer, converter)
        finally:
            # reset the unsupported args
            cls.unsupported_args = unsupported

        return convertible


@Converter.register(tf.keras.layers.Activation)
class ConvertActivation(LayerConverter):
    """Convert ``tf.keras.layers.Activation`` to Nengo objects."""

    def convert(self, node_id):
        output = self.add_nengo_obj(node_id, activation=self.layer.activation)

        self.add_connection(node_id, output)

        return output


@Converter.register(tf.keras.layers.Add)
class ConvertAdd(LayerConverter):
    """Convert ``tf.keras.layers.Add`` to Nengo objects."""

    def convert(self, node_id):
        output = self.add_nengo_obj(node_id)
        for i in range(len(self.layer.input)):
            self.add_connection(node_id, output, input_idx=i)

        return output


@Converter.register(tf.keras.layers.Average)
class ConvertAverage(LayerConverter):
    """Convert ``tf.keras.layers.Average`` to Nengo objects."""

    def convert(self, node_id):
        output = self.add_nengo_obj(node_id)
        for i in range(len(self.layer.input)):
            self.add_connection(
                node_id, output, input_idx=i, transform=1 / len(self.layer.input)
            )

        return output


@Converter.register(tf.keras.layers.AvgPool1D)
@Converter.register(tf.keras.layers.MaxPool1D)
@Converter.register(tf.keras.layers.GlobalAvgPool1D)
@Converter.register(tf.keras.layers.GlobalMaxPool1D)
class ConvertAvgPool1D(ConvertAvgPool):
    """
    Convert ``tf.keras.layers.AvgPool1D`` to Nengo objects.

    Also works for ``tf.keras.layers.GlobalAvgPool1D``, and
    ``tf.keras.layers.MaxPool1D``/``GlobalMaxPool1D`` (if ``max_to_avg_pool=True``).
    """

    def convert(self, node_id):
        return super().convert(node_id, dimensions=1)


@Converter.register(tf.keras.layers.AvgPool2D)
@Converter.register(tf.keras.layers.MaxPool2D)
@Converter.register(tf.keras.layers.GlobalAvgPool2D)
@Converter.register(tf.keras.layers.GlobalMaxPool2D)
class ConvertAvgPool2D(ConvertAvgPool):
    """
    Convert ``tf.keras.layers.AvgPool2D`` to Nengo objects.

    Also works for ``tf.keras.layers.GlobalAvgPool2D``, and
    ``tf.keras.layers.MaxPool2D``/``GlobalMaxPool2D`` (if ``max_to_avg_pool=True``).
    """

    def convert(self, node_id):
        return super().convert(node_id, dimensions=2)


@Converter.register(tf.keras.layers.AvgPool3D)
@Converter.register(tf.keras.layers.MaxPool3D)
@Converter.register(tf.keras.layers.GlobalAvgPool3D)
@Converter.register(tf.keras.layers.GlobalMaxPool3D)
class ConvertAvgPool3D(ConvertAvgPool):
    """
    Convert ``tf.keras.layers.AvgPool3D`` to Nengo objects.

    Also works for ``tf.keras.layers.GlobalAvgPool3D``, and
    ``tf.keras.layers.MaxPool3D``/``GlobalMaxPool3D`` (if ``max_to_avg_pool=True``).
    """

    def convert(self, node_id):
        return super().convert(node_id, dimensions=3)


@Converter.register(BatchNormalizationV1)
@Converter.register(BatchNormalizationV2)
class ConvertBatchNormalization(LayerConverter):
    """Convert ``tf.keras.layers.BatchNormalization`` to Nengo objects."""

    def convert(self, node_id):
        # look up the batch normalization parameters
        if self.layer.scale:
            gamma = tf.keras.backend.get_value(self.layer.gamma)
        else:
            gamma = 1

        if self.layer.center:
            beta = tf.keras.backend.get_value(self.layer.beta)
        else:
            beta = 0

        mean, variance = tf.keras.backend.batch_get_value(
            (self.layer.moving_mean, self.layer.moving_variance)
        )

        # compute the fixed affine transform values for this layer
        variance += self.layer.epsilon

        stddev = np.sqrt(variance)

        scale = gamma / stddev
        bias = beta - scale * mean

        # build output object
        output = self.add_nengo_obj(node_id)

        # the batch normalization parameters will be n-dimensional, where n is the
        # length of the axis specified in the batch normalization layer. so we need
        # to set up a connection structure so that all the elements of the output
        # corresponding to one of those axis elements will share the same parameter

        assert len(self.layer.axis) == 1
        assert self.layer.axis[0] > 0
        axis = self.layer.axis[0] - 1  # not counting batch dimension

        idxs = np.arange(output.size_in).reshape(self.output_shape(node_id))
        slices = [slice(None) for _ in range(len(idxs.shape))]

        # broadcast scale/bias along the non-axis dimensions, so that we can apply the
        # same scale/bias to all those elements
        broadcast_scale = np.zeros(self.output_shape(node_id))
        broadcast_bias = np.zeros(self.output_shape(node_id))
        for i in range(idxs.shape[axis]):
            slices[axis] = i
            broadcast_scale[tuple(slices)] = scale[i]
            broadcast_bias[tuple(slices)] = bias[i]
        broadcast_scale = np.ravel(broadcast_scale)
        broadcast_bias = np.ravel(broadcast_bias)

        # connect up bias node to output
        bias_node = nengo.Node(
            broadcast_bias, label="%s.%d.bias" % (self.layer.name, node_id)
        )
        conn = nengo.Connection(bias_node, output, synapse=None)
        self.set_trainable(conn, False)

        # connect input to output, scaled by the batch normalization scale
        self.add_connection(node_id, output, transform=broadcast_scale)

        return output

    @classmethod
    def convertible(cls, layer, converter):
        if not converter.inference_only and layer.trainable:
            msg = (
                "Cannot convert BatchNormalization layer to native Nengo objects "
                "unless inference_only=True or layer.trainable=False"
            )
            return False, msg

        return super().convertible(layer, converter)


@Converter.register(tf.keras.layers.Concatenate)
class ConvertConcatenate(LayerConverter):
    """Convert ``tf.keras.layers.Concatenate`` to Nengo objects."""

    def convert(self, node_id):
        output = self.add_nengo_obj(node_id)

        # axis-1 because not counting batch dimension
        axis = self.layer.axis - 1 if self.layer.axis > 0 else self.layer.axis

        idxs = np.arange(np.prod(self.output_shape(node_id))).reshape(
            self.output_shape(node_id)
        )
        slices = [slice(None) for _ in range(idxs.ndim)]
        offsets = np.cumsum([shape[axis] for shape in self.input_shape(node_id)])
        offsets = np.concatenate(([0], offsets))

        for i in range(len(self.layer.input)):
            slices[axis] = slice(offsets[i], offsets[i + 1])
            self.add_connection(
                node_id, output[np.ravel(idxs[tuple(slices)])], input_idx=i
            )

        return output

    @classmethod
    def convertible(cls, layer, converter):
        if layer.axis == 0:
            msg = "Cannot concatenate along batch dimension (axis 0)"
            return False, msg

        return super().convertible(layer, converter)


class ConvertConv(LayerConverter):
    """Base class for converting convolutional layers to Nengo objects."""

    has_weights = True

    def convert(self, node_id, dimensions):
        # look up parameter values from source layer
        if self.layer.use_bias:
            kernel, biases = tf.keras.backend.batch_get_value(
                (self.layer.kernel, self.layer.bias)
            )
        else:
            kernel = tf.keras.backend.get_value(self.layer.kernel)
            biases = None

        # create nengo object to implement activation function
        output = self.add_nengo_obj(node_id, activation=self.layer.activation)

        if self.layer.use_bias:
            # conv layer biases are per-output-channel, rather than per-output-element,
            # so we need to set up a nengo connection structure that will have one
            # bias parameter shared across all the spatial dimensions

            # add trainable bias weights
            bias_node = nengo.Node([1], label="%s.%d.bias" % (self.layer.name, node_id))
            bias_relay = nengo.Node(
                size_in=len(biases),
                label="%s.%d.bias_relay" % (self.layer.name, node_id),
            )
            nengo.Connection(
                bias_node, bias_relay, transform=biases[:, None], synapse=None
            )

            # use a non-trainable sparse transform to broadcast biases along all
            # non-channel dimensions
            broadcast_indices = []
            idxs = np.arange(np.prod(self.output_shape(node_id))).reshape(
                self.output_shape(node_id)
            )
            slices = [slice(None) for _ in range(len(self.output_shape(node_id)))]
            n_spatial = np.prod(
                self.output_shape(node_id)[:-1]
                if self.layer.data_format == "channels_last"
                else self.output_shape(node_id)[1:]
            )
            axis = -1 if self.layer.data_format == "channels_last" else 0
            for i in range(self.output_shape(node_id)[axis]):
                slices[axis] = i
                broadcast_indices.extend(
                    tuple(zip(np.ravel(idxs[tuple(slices)]), [i] * n_spatial))
                )
            conn = nengo.Connection(
                bias_relay,
                output,
                transform=nengo.Sparse(
                    (output.size_in, bias_relay.size_out), indices=broadcast_indices
                ),
                synapse=None,
            )
            self.set_trainable(conn, False)

        # set up a convolutional transform that matches the layer parameters
        transform = nengo.Convolution(
            n_filters=self.layer.filters,
            input_shape=self.input_shape(node_id),
            kernel_size=self.layer.kernel_size,
            strides=self.layer.strides,
            padding=self.layer.padding,
            channels_last=self.layer.data_format == "channels_last",
            init=kernel,
        )

        self.add_connection(node_id, output, transform=transform, trainable=True)

        return output


@Converter.register(tf.keras.layers.Conv1D)
class ConvertConv1D(ConvertConv):
    """Convert ``tf.keras.layers.Conv1D`` to Nengo objects."""

    unsupported_args = [
        ("dilation_rate", (1,)),
    ]
    unsupported_training_args = [
        "kernel_regularizer",
        "bias_regularizer",
        "activity_regularizer",
        "kernel_constraint",
        "bias_constraint",
    ]

    def convert(self, node_id):
        return super().convert(node_id, dimensions=1)


@Converter.register(tf.keras.layers.Conv2D)
class ConvertConv2D(ConvertConv):
    """Convert ``tf.keras.layers.Conv2D`` to Nengo objects."""

    unsupported_args = [
        ("dilation_rate", (1, 1)),
    ]
    unsupported_training_args = [
        "kernel_regularizer",
        "bias_regularizer",
        "activity_regularizer",
        "kernel_constraint",
        "bias_constraint",
    ]

    def convert(self, node_id):
        return super().convert(node_id, dimensions=2)


@Converter.register(tf.keras.layers.Conv3D)
class ConvertConv3D(ConvertConv):
    """Convert ``tf.keras.layers.Conv3D`` to Nengo objects."""

    unsupported_args = [
        ("dilation_rate", (1, 1, 1)),
    ]
    unsupported_training_args = [
        "kernel_regularizer",
        "bias_regularizer",
        "activity_regularizer",
        "kernel_constraint",
        "bias_constraint",
    ]

    def convert(self, node_id):
        return super().convert(node_id, dimensions=3)


@Converter.register(tf.keras.layers.Dense)
class ConvertDense(LayerConverter):
    """Convert ``tf.keras.layers.Dense`` to Nengo objects."""

    unsupported_training_args = [
        "kernel_regularizer",
        "bias_regularizer",
        "activity_regularizer",
        "kernel_constraint",
        "bias_constraint",
    ]

    has_weights = True

    def convert(self, node_id):
        # look up parameter values from source layer
        if self.layer.use_bias:
            weights, biases = tf.keras.backend.batch_get_value(
                (self.layer.kernel, self.layer.bias)
            )
        else:
            weights = tf.keras.backend.get_value(self.layer.kernel)
            biases = None

        # create nengo object to implement activation function and biases
        output = self.add_nengo_obj(
            node_id, activation=self.layer.activation, biases=biases
        )

        # add connection to implement the dense weights
        self.add_connection(node_id, output, transform=weights.T, trainable=True)

        return output


@Converter.register(tf.keras.layers.Flatten)
class ConvertFlatten(LayerConverter):
    """Convert ``tf.keras.layers.Flatten`` to Nengo objects."""

    def convert(self, node_id):
        # noop, same as reshape
        return self.get_input_obj(node_id)


@Converter.register(tf.keras.layers.InputLayer)
class ConvertInput(LayerConverter):
    """Convert ``tf.keras.layers.InputLayer`` to Nengo objects."""

    def convert(self, node_id):
        shape = self.output_shape(node_id)
        if any(x is None for x in shape):
            raise ValueError(
                "Input shapes must be fully specified; got %s. If inputs contain "
                "`None` in the first axis to indicate a variable number of timesteps, "
                "set `temporal_model=True` on the `Converter`." % (shape,)
            )
        output = nengo.Node(size_in=np.prod(shape), label=self.layer.name)

        logger.info("Created %s", output)

        return output


@Converter.register(tf.keras.layers.ReLU)
class ConvertReLU(LayerConverter):
    """Convert ``tf.keras.layers.ReLU`` to Nengo objects."""

    unsupported_args = ["max_value", ("threshold", 0)]

    def convert(self, node_id):
        if self.layer.negative_slope == 0:
            activation = tf.nn.relu
        else:
            activation = LeakyReLU(negative_slope=self.layer.negative_slope)

        output = self.add_nengo_obj(node_id, biases=None, activation=activation)

        self.add_connection(node_id, output)

        return output


@Converter.register(tf.keras.layers.LeakyReLU)
class ConvertLeakyReLU(LayerConverter):
    """Convert ``tf.keras.layers.LeakyReLU`` to Nengo objects."""

    def convert(self, node_id):
        output = self.add_nengo_obj(
            node_id, biases=None, activation=LeakyReLU(negative_slope=self.layer.alpha)
        )

        self.add_connection(node_id, output)

        return output


@Converter.register(tf.keras.layers.Reshape)
class ConvertReshape(LayerConverter):
    """Convert ``tf.keras.layers.Reshape`` to Nengo objects."""

    def convert(self, node_id):
        # nengo doesn't pass shape information between objects (everything is just a
        # vector), so we don't actually need to do anything here, we just return
        # the input layer. layers that require shape information can look it up from
        # the input_shape attribute of their layer
        return self.get_input_obj(node_id)


class ConvertUpSampling(LayerConverter):
    """Base class for converting upsampling layers to Nengo objects."""

    def convert(self, node_id, dimensions):
        output = self.add_nengo_obj(node_id)

        channels_last = (
            getattr(self.layer, "data_format", "channels_last") == "channels_last"
        )
        reps = np.atleast_1d(self.layer.size)

        input_shape = self.input_shape(node_id)
        input_shape = input_shape[:-1] if channels_last else input_shape[1:]

        output_shape = self.output_shape(node_id)
        output_shape = output_shape[:-1] if channels_last else output_shape[1:]

        # figure out nearest neighbour (in the source array) for each point in the
        # upsampled array
        input_pts = np.stack(
            np.meshgrid(
                *[
                    np.arange(0, x * reps[i], reps[i], dtype=np.float32)
                    for i, x in enumerate(input_shape)
                ],
                indexing="ij",
            ),
            axis=-1,
        )
        input_pts += (reps - 1) / 2  # shift to centre of upsampled block
        input_pts = np.reshape(input_pts, (-1, dimensions))

        upsampled_pts = np.stack(
            np.meshgrid(
                *[np.arange(x) for x in output_shape],
                indexing="ij",
            ),
            axis=-1,
        )
        upsampled_pts = np.reshape(upsampled_pts, (-1, dimensions))

        dists = np.linalg.norm(
            input_pts[..., None] - upsampled_pts.T[None, ...], axis=1
        )
        nearest = np.argmin(dists, axis=0)

        # duplicate along channel axis
        idxs = np.arange(np.prod(self.input_shape(node_id))).reshape(
            self.input_shape(node_id)
        )
        if channels_last:
            idxs = np.reshape(idxs, (-1, idxs.shape[-1]))
            idxs = idxs[nearest]
        else:
            idxs = np.reshape(idxs, (idxs.shape[0], -1))
            idxs = idxs[:, nearest]

        # connect from pre, using idxs to perform upsampling
        pre = self.get_input_obj(node_id, tensor_idx=0)[np.ravel(idxs)]
        conn = nengo.Connection(
            pre,
            output,
            synapse=self.converter.synapse
            if isinstance(pre, nengo.ensemble.Neurons)
            else None,
        )
        self.set_trainable(conn, False)

        return output


@Converter.register(tf.keras.layers.UpSampling1D)
class ConvertUpSampling1D(ConvertUpSampling):
    """Convert ``tf.keras.layers.UpSampling1D`` to Nengo objects."""

    def convert(self, node_id):
        return super().convert(node_id, dimensions=1)


@Converter.register(tf.keras.layers.UpSampling2D)
class ConvertUpSampling2D(ConvertUpSampling):
    """Convert ``tf.keras.layers.UpSampling2D`` to Nengo objects."""

    unsupported_args = [("interpolation", "nearest")]

    def convert(self, node_id):
        return super().convert(node_id, dimensions=2)


@Converter.register(tf.keras.layers.UpSampling3D)
class ConvertUpSampling3D(ConvertUpSampling):
    """Convert ``tf.keras.layers.UpSampling3D`` to Nengo objects."""

    def convert(self, node_id):
        return super().convert(node_id, dimensions=3)


class ConvertZeroPadding(LayerConverter):
    """Base class for converting zero-padding layers to Nengo objects."""

    def convert(self, node_id, dimensions):
        output = self.add_nengo_obj(node_id)

        # zeropadding1d doesn't doesn't have data_format, assumes channels_last
        channels_first = (
            getattr(self.layer, "data_format", "channels_last") == "channels_first"
        )

        # the strategy here is that we'll create a nengo node of the full padded size,
        # and then connect up the input to the subset of those node elements
        # corresponding to the inner, non-padded elements. so we need to figure out
        # what the indices are that we need to connect to.

        # build slices representing the non-padded elements within the output shape
        slices = []
        if channels_first:
            slices.append(slice(None))
        for i in range(dimensions):
            if dimensions == 1:
                top_pad, bottom_pad = self.layer.padding
            else:
                top_pad, bottom_pad = self.layer.padding[i]
            length = self.output_shape(node_id)[i + channels_first]
            slices.append(slice(top_pad, length - bottom_pad))

        # apply slices to index array to get the list of indices we want to connect to
        idxs = np.arange(output.size_in).reshape(self.output_shape(node_id))
        idxs = np.ravel(idxs[tuple(slices)])

        # connect up the input to the appropriate indices
        self.add_connection(node_id, output[idxs])

        return output


@Converter.register(tf.keras.layers.ZeroPadding1D)
class ConvertZeroPadding1D(ConvertZeroPadding):
    """Convert ``tf.keras.layers.ZeroPadding1D`` to Nengo objects."""

    def convert(self, node_id):
        return super().convert(node_id, dimensions=1)


@Converter.register(tf.keras.layers.ZeroPadding2D)
class ConvertZeroPadding2D(ConvertZeroPadding):
    """Convert ``tf.keras.layers.ZeroPadding2D`` to Nengo objects."""

    def convert(self, node_id):
        return super().convert(node_id, dimensions=2)


@Converter.register(tf.keras.layers.ZeroPadding3D)
class ConvertZeroPadding3D(ConvertZeroPadding):
    """Convert ``tf.keras.layers.ZeroPadding3D`` to Nengo objects."""

    def convert(self, node_id):
        return super().convert(node_id, dimensions=3)


class ConvertKerasSpiking(LayerConverter):
    """Base class for converting KerasSpiking layers."""

    unsupported_args = [
        ("return_state", False),
        ("return_sequences", True),
        ("time_major", False),
    ]

    allow_fallback = False

    def convert(self, node_id):
        if self.layer.dt != 0.001:
            # this is the default keras-spiking dt; warn if it has been changed, as
            # those changes will be overridden by sim.dt. we don't want to error,
            # because this isn't necessarily a problem (they could set sim.dt to
            # match the layer dt).
            # TODO: add some kind of callback to check that sim.dt matches layer.dt?
            warnings.warn(
                "Ignoring %s.dt=%f parameter; dt will be controlled by Simulator.dt"
                % (type(self.layer).__name__, self.layer.dt)
            )
        if self.layer.stateful:
            warnings.warn(
                "Ignoring %s.stateful=True parameter; statefulness will "
                "be controlled by Simulator" % type(self.layer).__name__
            )


@Converter.register(compat.SpikingActivation)
class ConvertSpikingActivation(ConvertKerasSpiking):
    """Convert ``keras_spiking.SpikingActivation`` to Nengo objects."""

    unsupported_training_args = [("spiking_aware_training", False)]

    def convert(self, node_id):
        super().convert(node_id)

        # note: not applying swap_activations to these activations, because that's
        # likely to be confusing (e.g. swapping relu to spiking relu, and then wrapping
        # that in regularspiking by accident)
        activation = self.activation_map.get(self.layer.activation, None)

        if activation is None:
            # TODO: allow fallback within SpikingActivation?
            raise TypeError(
                "SpikingActivation activation type (%s) does not have a native Nengo "
                "equivalent" % self.layer.activation
            )

        initial_state = tf.keras.backend.get_value(
            self.layer.layer.cell.get_initial_state(batch_size=1, dtype="float32")
        )[0]

        activation = nengo.RegularSpiking(
            activation, initial_state={"voltage": initial_state}
        )

        output = self.add_nengo_obj(node_id, biases=None, activation=activation)

        self.add_connection(node_id, output)

        return output

    @classmethod
    def convertible(cls, layer, converter):
        if version.parse(nengo.__version__) <= version.parse("3.0.0"):
            return False, "Converting SpikingActivation layers requires Nengo>=3.1.0"

        return super().convertible(layer, converter)


@Converter.register(compat.Lowpass)
@Converter.register(compat.Alpha)
class ConvertLowpassAlpha(ConvertKerasSpiking):
    """Convert ``keras_spiking.Lowpass`` to Nengo objects."""

    def convert(self, node_id):
        super().convert(node_id)

        output = self.add_nengo_obj(node_id)

        tau = np.ravel(tf.keras.backend.get_value(self.layer.layer.cell.tau_var))[0]

        synapse = (
            nengo.Lowpass(tau)
            if isinstance(self.layer, compat.Lowpass)
            else nengo.Alpha(tau)
        )

        self.add_connection(node_id, output, synapse=synapse)

        return output

    @classmethod
    def convertible(cls, layer, converter):
        if not converter.inference_only and layer.trainable:
            msg = (
                "Cannot convert %s layer to native Nengo objects "
                "unless inference_only=True or layer.trainable=False"
                % type(layer).__name__
            )
            return False, msg

        if not np.allclose(
            tf.keras.backend.get_value(layer.layer.cell.initial_level), 0
        ):
            msg = (
                "Cannot convert a %s layer to native Nengo objects with "
                "initial_level != 0 (this probably means that training has been "
                "applied to the layer before conversion)" % type(layer).__name__
            )
            return False, msg

        tau = tf.keras.backend.get_value(layer.layer.cell.tau_var)
        if not np.allclose(tau, np.ravel(tau)[0]):
            msg = (
                "Cannot convert a %s layer to native Nengo objects with different "
                "tau values for each element (this probably means that training "
                "has been applied to the layer before conversion)"
                % type(layer).__name__
            )
            return False, msg

        return super().convertible(layer, converter)


@Converter.register(tf.keras.layers.TimeDistributed)
class ConvertTimeDistributed(LayerConverter):
    """Convert ``tf.keras.layers.TimeDistributed`` to Keras objects."""

    allow_fallback = False

    def convert(self, node_id):
        # nengo models are already temporal, so we don't need to do anything special
        # to make the wrapped layer temporal, we just convert the wrapped layer

        # for some reason, using a layer inside a TimeDistributed wrapper doesn't
        # update it's inbound_nodes. so we'll call the wrapped layer on the wrapper's
        # input. note that the wrapper's input has the extra time dimension, so we slice
        # out just the first timestep (it doesn't matter what the values are)
        input_node = self.layer.inbound_nodes[node_id]
        sliced_input = input_node.input_tensors[:, 0]
        _ = self.layer.layer(sliced_input)

        # add the sliced input we created above to the layer map so that when
        # we look up the input nengo objects of the wrapped layer we get the input of
        # the timedistributed layer
        (
            input_layer,
            input_node_id,
            _,
        ) = input_node.input_tensors._keras_history
        (
            sliced_input_layer,
            sliced_input_node_id,
            _,
        ) = sliced_input._keras_history
        layer_map = nengo.Network.context[-1].layer_map
        layer_map[sliced_input_layer][sliced_input_node_id] = layer_map[input_layer][
            input_node_id
        ]

        # temporarily set temporal_model=False so that the wrapped layer builds
        # correctly
        self.converter.temporal_model = False
        try:
            result = self.converter.get_converter(self.layer.layer).convert(
                # use the nodeid of the just created node
                len(self.layer.layer.inbound_nodes)
                - 1
            )
        finally:
            self.converter.temporal_model = True

        return result

    @classmethod
    def convertible(cls, layer, converter):
        if not converter.temporal_model:
            return (
                False,
                "TimeDistributed layers can only be converted when temporal_model=True",
            )

        return super().convertible(layer, converter)
